/*
 * c1.q.c
 * Copyright (C) 2008-2016, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>
#include <string.h>

#include <c1-defs.h>
#include <c1-inter.h>
#include <c1-types.h>
#include <e4.h>
#include <exerrors.h>

#define false(e)			0

typedef union link_type {
    X1f4_E4_C_MODE mode;
    X1f4_E4_C_REAL real;
    X1f4_E4_C_TEXT text;
    X1f4_E4_C_USER user;
} link_type;

extern const void *const x1f4_a1_walk_link;

static int case_back(struct c1_trans19_type *, struct c1_node_type *);
static int case_miss(struct c1_trans19_type *, struct c1_node_type *);
static int case_node(struct c1_trans19_type *, struct c1_node_type *);
static int case_slip(struct c1_trans19_type *, struct c1_node_type *, int);
static int copy_node(struct c1_trans19_type *, struct c1_node_type *,
		     int *);
static int ever_node(struct c1_trans19_type *, struct c1_node_type *,
		     int *);
static int fine_last(struct c1_trans19_type *, struct x1f4_transfer_type *,
		     int *);
static int fine_link(struct c1_trans19_type *, struct x1f4_transfer_type *,
		     int *, int (*) (void *, void *, void **), unsigned);
static int head_node(struct c1_trans19_type *, struct c1_node_type *);
static int lead_back(struct c1_trans19_type *, struct c1_node_type *);
static int lead_miss(struct c1_trans19_type *, struct c1_node_type *);
static int lead_node(struct c1_trans19_type *, struct c1_node_type *);
static int long_lock(struct c1_trans19_type *, void *, void *);
static int long_port(struct c1_trans19_type *, struct c1_program_type *, int,
		     void *, struct x1f4_dispatch_type **);
static int lose_node(struct c1_trans19_type *, struct c1_node_type *);
static int pick_lead(struct c1_trans19_type *, int);
static int pick_slip(struct c1_trans19_type *, int);
static int post_lead(struct c1_trans19_type *);
static int slip_node(struct c1_trans19_type *);
static int tail_node(struct c1_trans19_type *, struct c1_node_type *);

static int
case_back(struct c1_trans19_type *trans19_data,
	  struct c1_node_type *node_data)
{
    int status;

    if (0) {
    } else {
	int slipping = 0;

	status = copy_node(trans19_data, node_data, &slipping);
	if (status) {
	} else {
	    if (slipping) {
	    } else {
		status = tail_node(trans19_data, node_data + 1);
	    }
	}
    }

    return status;
}


static int
case_miss(struct c1_trans19_type *trans19_data,
	  struct c1_node_type *node_data)
{
    int status;

    status = head_node(trans19_data, node_data + 1);
    if (status) {
    } else {
	int slipping = 0;

	status = copy_node(trans19_data, node_data, &slipping);
	if (status) {
	} else {
	    if (slipping) {
	    } else {
		status = tail_node(trans19_data, node_data + 1);
	    }
	}
    }

    return status;
}


static int
case_node(struct c1_trans19_type *trans19_data,
	  struct c1_node_type *node_data)
{
    int status;

    if (node_data->state & Fst_STATE) {
	status = case_miss(trans19_data, node_data);
    } else {
	int deck;

	status = x1f4_cast_expression
	    ((node_data + 1)->data.data, &deck, trans19_data->i.miss,
	     trans19_data->i.used);
	if (status) {
	} else {
	    case_slip(trans19_data, node_data, deck);
	}
    }

    return status;
}


static int
case_slip(struct c1_trans19_type *trans19_data,
	  struct c1_node_type *node_data, int deck)
{
    if (deck) {
	trans19_data->e.slip = node_data->node_data;
    } else {
	node_data++;
	trans19_data->e.slip = node_data->node_data->node_data;
    }

    return 0;
}


static int
copy_node(struct c1_trans19_type *trans19_data,
	  struct c1_node_type *node_data, int *sideline)
{
    int status;
    void *expression, *slip;

    expression = (node_data + 1)->data.data;

    slip = trans19_data->e.port;

    while (1) {
	int slipping;
	struct x1f4_transfer_type transfer;
	union link_type link;

	status = x1f4_span_expression
	    (expression, slip, 0, &slipping, &link, &transfer,
	     trans19_data->i.used);
	if (status) {
	    break;
	} else {
	    if (slipping) {
		status = fine_last(trans19_data, &transfer, sideline);
		if (*sideline) {
		    break;
		} else {
		    if (status) {
			break;
		    }
		}
	    } else {
		int type;

		type = x1f4_type_expression(expression);
		if (type == X1f4_E4_MODE) {
		    type = *(X1f4_E4_C_MODE *) &link;
		} else if (type == X1f4_E4_REAL) {
		    type = *(X1f4_E4_C_REAL *) &link ? 1 : 0;
		} else {
		    type = 1;
		}

		if (1) {
		    case_slip(trans19_data, node_data, type);
		    if (1) {
			*sideline = 0;
			if (1) {
			    break;
			}
		    }
		}
	    }
	}
    }

    return status;
}


static int
ever_node(struct c1_trans19_type *trans19_data,
	  struct c1_node_type *node_data, int *sideline)
{
    int status;
    void *expression, *slip;

    expression = (node_data + 0)->data.data;

    slip = trans19_data->e.port;

    while (1) {
	int slipping;
	struct x1f4_transfer_type transfer;
	union link_type link;

	status = x1f4_span_expression
	    (expression, slip, 0, &slipping, &link, &transfer,
	     trans19_data->i.used);
	if (status) {
	    break;
	} else {
	    if (slipping) {
		status = fine_last(trans19_data, &transfer, sideline);
		if (*sideline) {
		    break;
		} else {
		    if (status) {
			break;
		    }
		}
	    } else {
		if (1) {
		    trans19_data->e.slip = node_data->node_data;
		    if (1) {
			*sideline = 0;
			if (1) {
			    break;
			}
		    }
		}
	    }
	}
    }

    return status;
}


/*
 * NOTES
 *
 * the function does execute recursively.  If the WALK_LINK function(s)
 * recurse(s) for too many times, the stack will be overrun.  The condition can
 * (!?) be cured, and yet the functions are application introduced, and curing
 * application code is beyond scope here.
 */
static int
fine_last(struct c1_trans19_type *trans19_data,
	  struct x1f4_transfer_type *transfer_data, int *sideline)
{
    int (*function) (void *, void *, void **), status;
    const struct x1f4_function_type *function_data;

    function_data = transfer_data->function_data;

    function = function_data->function;

    if (0) {
    } else {
	if (0) {
	} else {
	    if (function) {
		unsigned flags;

		flags = function_data->flags;
		if (flags & (X1f4_E4_LEAD_CALL | X1f4_E4_WALK_LINK)) {
		    if (flags & X1f4_E4_WALK_LINK) {
			status = fine_link
			    (trans19_data, transfer_data, sideline, function,
			     flags);
		    } else {
			trans19_data->e.sample = 1;
			trans19_data->x.transfer = *transfer_data;
			*sideline = 1;
			if (1) {
			    status = 0;
			}
		    }
		} else {
		    void **input, *output;

		    input = transfer_data->input;
		    output = transfer_data->output;

		    input[-1] = trans19_data->i.used;

		    status = function(transfer_data->context, output, input);
		}
	    } else {
		status = _x1f4_c1_stat_miss(trans19_data->f.program_data);
	    }
	}
    }

    return status;
}


static int
fine_link(struct c1_trans19_type *trans19_data,
	  struct x1f4_transfer_type *transfer_data, int *sideline,
	  int (*function) (void *, void *, void **), unsigned flags)
{
    int status;
    void *mid[3];

    mid[0] = transfer_data->output;
    mid[1] = transfer_data->input;

    transfer_data->input[-1] = trans19_data->i.used;

    status = function(transfer_data->context, (void *) x1f4_a1_walk_link, mid);
    if (status) {
    } else {
	struct x1f4_dispatch_type *dispatch_data;

	dispatch_data = mid[2];
	while (dispatch_data) {
	    status = fine_last
		(trans19_data, &dispatch_data->transfer, sideline);
	    if (*sideline) {
		if (1) {
		    status = long_lock
			(trans19_data, dispatch_data, transfer_data->output);
		} else {
		}

		break;
	    } else {
		if (status) {
		    if (status == X1f4_EX_CAN_CONTINUE) {
		    } else {
			break;
		    }
		}
		{
		    status = dispatch_data->call
			(dispatch_data->back, status, transfer_data->output,
			 &dispatch_data);
		    if (status) {
			if (status == X1f4_EX_CAN_CONTINUE) {
			    dispatch_data = NULL;
			} else {
			    break;
			}
		    }
		}
	    }
	}
    }

    return status;
}


static int
head_node(struct c1_trans19_type *trans19_data,
	  struct c1_node_type *node_data)
{
    return x1f4_land_expression
	(node_data->data.data, trans19_data->i.miss, &trans19_data->e.port);
}


static int
lead_back(struct c1_trans19_type *trans19_data,
	  struct c1_node_type *node_data)
{
    int status;

    if (0) {
    } else {
	int slipping = 0;

	status = ever_node(trans19_data, node_data, &slipping);
	if (status) {
	} else {
	    if (slipping) {
	    } else {
		status = tail_node(trans19_data, node_data + 0);
	    }
	}
    }

    return status;
}


static int
lead_miss(struct c1_trans19_type *trans19_data,
	  struct c1_node_type *node_data)
{
    int status;

    status = head_node(trans19_data, node_data + 0);
    if (status) {
    } else {
	int slipping = 0;

	status = ever_node(trans19_data, node_data, &slipping);
	if (status) {
	} else {
	    if (slipping) {
	    } else {
		status = tail_node(trans19_data, node_data + 0);
	    }
	}
    }

    return status;
}


static int
lead_node(struct c1_trans19_type *trans19_data,
	  struct c1_node_type *node_data)
{
    int status;

    if (node_data->state & Fst_STATE) {
	status = lead_miss(trans19_data, node_data);
    } else {
	status = x1f4_last_expression
	    (node_data->data.data, trans19_data->i.miss,
	     trans19_data->i.used);
	if (status) {
	} else {
	    if (0) {
	    } else {
		trans19_data->e.slip = node_data->node_data;
	    }
	}
    }

    return status;
}


static int
long_lock(struct c1_trans19_type *trans19_data, void *dispatch,
	  void *dispatched_output)
{
    int status;
    struct c1_program_type *program_data;
    void *fast;

    program_data = trans19_data->f.program_data;

    status = program_data->n.link
	(program_data->n.data, &fast, sizeof(struct c1_fast_type));
    if (status) {
	status = _x1f4_c1_stat_link(program_data);
    } else {
	struct c1_fast_type *fast_data, *fast_tail;

	fast_data = fast;

	fast_data->fast_data = NULL;
	fast_data->dispatch_data = dispatch;
	fast_data->dispatched_output = dispatched_output;

	fast_tail = trans19_data->e.tail;

	trans19_data->e.tail = fast_data;

	if (fast_tail) {
	    fast_tail->fast_data = fast_data;
	} else {
	    trans19_data->e.head = fast_data;
	}
    }

    return status;
}


static int
long_port(struct c1_trans19_type *trans19_data,
	  struct c1_program_type *program_data, int delete, void *output,
	  struct x1f4_dispatch_type **dispatch)
{
    int status;
    struct x1f4_dispatch_type *dispatch_data;

    dispatch_data = *dispatch;

    while (1) {
	delete = dispatch_data->call
	    (dispatch_data->back, delete, output, &dispatch_data);
	if (delete) {
	    dispatch_data = NULL;
	    if (delete == X1f4_EX_CAN_CONTINUE) {
	    } else {
		status = delete;
		if (1) {
		    break;
		}
	    }
	}
	{
	    if (dispatch_data) {
		int slipping = 0;

		delete = fine_last
		    (trans19_data, &dispatch_data->transfer, &slipping);
		if (delete) {
		    if (delete == X1f4_EX_CAN_CONTINUE) {
			/*
			 * no aime function failed (the aime function executors
			 * will not return CAN CONTINUE for failure)
			 */
		    } else {
			dispatch_data = NULL;
			status = delete;
			break;
		    }
		} else {
		    if (slipping) {
			status = 0;
			if (1) {
			    break;
			}
		    }
		}
	    } else {
		if (0) {
		} else {
		    status = delete;
		    if (1) {
			break;
		    }
		}
	    }
	}
    }

    *dispatch = dispatch_data;

    return status;
}


static int
lose_node(struct c1_trans19_type *trans19_data,
	  struct c1_node_type *node_data)
{
    int status;

    if (node_data->data.data) {
	status = lead_back(trans19_data, node_data);
    } else {
	status = case_back(trans19_data, node_data);
    }

    return status;
}


static int
pick_lead(struct c1_trans19_type *trans19_data, int delete)
{
    int status;

    do {
	struct c1_fast_type *fast_data;
	struct c1_program_type *program_data;
	struct x1f4_dispatch_type *dispatch_data;
	void *output;

	program_data = trans19_data->f.program_data;

	fast_data = trans19_data->e.head;

	dispatch_data = fast_data->dispatch_data;
	output = fast_data->dispatched_output;

	status = long_port
	    (trans19_data, program_data, delete, output, &dispatch_data);
	delete = status;
	if (status == X1f4_EX_CAN_CONTINUE) {
	    status = 0;
	}
	if (status) {
	    break;
	} else {
	    if (dispatch_data) {
		break;
	    } else {
		trans19_data->e.head = fast_data->fast_data;
		if (trans19_data->e.head) {
		} else {
		    trans19_data->e.tail = NULL;
		}

		status = program_data->n.free
		    (program_data->n.data, fast_data);
		if (status) {
		    status = _x1f4_c1_stat_free(program_data);
		    if (1) {
			break;
		    }
		}

		if (trans19_data->e.head) {
		} else {
		    if (delete) {
			status = delete;
		    } else {
			status = lose_node(trans19_data, trans19_data->e.slip);
		    }

		    break;
		}
	    }
	}
    } while (1);

    return status;
}


static int
pick_slip(struct c1_trans19_type *trans19_data, int status)
{
    while (status == X1f4_EX_CAN_CONTINUE) {
	if (trans19_data->e.port) {
	    int excess;

	    excess = _x1f4_c1_miss_port(trans19_data);
	    if (excess) {
		if (false(excess == X1f4_EX_CAN_CONTINUE)) {
		} else {
		    status = excess;
		    if (1) {
			break;
		    }
		}
	    } else {
		trans19_data->e.port = NULL;
	    }
	}

	if (0) {
	} else {
	    break;
	}
    }

    return status;
}


static int
post_lead(struct c1_trans19_type *trans19_data)
{
    int status;
    struct x1f4_transfer_type *transfer_data;

    trans19_data->e.sample = 0;

    transfer_data = &trans19_data->x.transfer;
    status = transfer_data->function_data->function
	(transfer_data->context, transfer_data->output, transfer_data->input);
    if (trans19_data->e.head) {
	status = pick_lead(trans19_data, status);
    } else {
	if (status) {
	} else {
	    status = lose_node(trans19_data, trans19_data->e.slip);
	}
    }

    return status;
}


static int
slip_node(struct c1_trans19_type *trans19_data)
{
    int status;
    struct c1_node_type *node_data;

    node_data = trans19_data->e.slip;
    if (node_data->data.data) {
	status = lead_node(trans19_data, node_data);
    } else {
	if ((node_data + 1)->data.data) {
	    status = case_node(trans19_data, node_data);
	} else {
	    node_data++;
	    trans19_data->e.slip = node_data->node_data;
	    status = 0;
	}
    }

    return status;
}


static int
tail_node(struct c1_trans19_type *trans19_data,
	  struct c1_node_type *node_data)
{
    int status;

    status = x1f4_tail_expression(node_data->data.data, &trans19_data->e.port);
    if (status) {
    } else {
	trans19_data->e.port = NULL;
    }

    return status;
}


#if 0
int
x1f4_slip_routine(struct c1_trans19_type *trans19_data)
{
    int status;

    if (trans19_data->e.sample) {
	status = post_lead(trans19_data);
    } else {
	if (1) {
	    status = slip_node(trans19_data);
	} else {
	}
    }
    if (status) {
	status = pick_slip(trans19_data, status);
    }

    return status;
}
#endif				/* 0 */


int
x1f4_step_routine(struct c1_trans19_type *trans19_data)
{
    int status;

    if (!trans19_data->e.slip) {
	status = X1f4_EX_COMPLETED;
    } else {
	if (trans19_data->e.sample) {
	    status = post_lead(trans19_data);
	} else {
	    if (1) {
		status = slip_node(trans19_data);
	    } else {
	    }
	}
	if (status) {
	    status = pick_slip(trans19_data, status);
	}
	if (status) {
	    status = -1;
	} else {
	    status = trans19_data->e.sample ? X1f4_EX_LEAD_CALL : 0;
	}
    }

    return status;
}
