/*
 * lxpart-a.0.c
 * Copyright (C) 2008-2015, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>
#include <string.h>

#include <e4-m.0.h>
#include <e4.h>
#include <lxcall.h>
#include <lxpart-defs.h>
#include <lxpart-inter.h>
#include <lxpart-types.h>

#define MAKE_SINGLE(a, b)		a

#define cxflat(cxflat) \
    ((struct cxpart_type *) (cxflat))

static int init_part(struct lxpart_type *, unsigned,
		     const struct screen_type *);
static int line_data(struct lxpart_type *, unsigned,
		     const struct screen_type *, void *, void *, void **);
static int line_fail(struct lxpart_type *, unsigned, void *, void *);
static int line_part(struct lxpart_type *, unsigned,
		     const struct screen_type *, void *);
static int line_sail(struct lxpart_type *, const struct screen_type *,
		     struct lxtext_type *, struct cxpart_type *);
static int link_part(struct lxpart_type *, unsigned,
		     const struct screen_type *);
static int pick_miss(unsigned *, unsigned *, const struct screen_type *);

static void line_sset(u_case_args_____0);

static const struct line_type ever_line[] = {
/* *INDENT-OFF* */
#define f_set_count			2
    {	{	"?_set",
		_libx1f4i0_lxpart_s_forward,	0,
		NULL,				f_set_count,
		X1f4_E4_KEEP_CALL | X1f4_E4_POST_TYPE
		| X1f4_E4_TEXT_LINK,
						5		},
	line_sset,							}
#define f_set_reach			(0 + (f_set_count << 1))
/* *INDENT-ON* */
#define ever_line_count			f_set_reach
};
static const struct x1f4_operator_type side_o[] = {
/* *INDENT-OFF* */
    {	MAKE_SINGLE("=", " "),  NULL,		0400,
	0,			NULL,
	X1f4_E4_BACK_LINK | X1f4_E4_E2ND_LINK | X1f4_E4_LEFT_XSET,
				1,
	NULL,			NULL					}
/* *INDENT-ON* */
};

static int
init_part(struct lxpart_type *lxpart_data, unsigned bits,
	  const struct screen_type *screen_data)
{
    int status;

    if (bits & AUTOLINK_LINK) {
	lxpart_data->link_v = screen_data->lock_v;
    } else {
	lxpart_data->link_v.data = NULL;
	lxpart_data->link_v.free = NULL;
	lxpart_data->link_v.link = NULL;
	lxpart_data->link_v.mode = NULL;
	lxpart_data->link_v.pick = NULL;
	lxpart_data->link_v.slip = NULL;
    }

    if (bits & CODELINK_LINK) {
	lxpart_data->link_w = screen_data->lock_w;
    } else {
	lxpart_data->link_w.data = NULL;
	lxpart_data->link_w.free = _x1f4_e4_free_data;
	lxpart_data->link_w.link = _x1f4_e4_link_data;
	lxpart_data->link_w.mode = _x1f4_e4_mode_data;
    }

    if (bits & MISSBAIL_LINK) {
	lxpart_data->link_r.call = screen_data->link_r.call;
	lxpart_data->link_r.fine = screen_data->link_r.fine;
	lxpart_data->link_r.miss = screen_data->link_r.miss;
	lxpart_data->link_r.text = screen_data->link_r.text;
    } else {
	lxpart_data->link_r.call = NULL;
	lxpart_data->link_r.fine = NULL;
	lxpart_data->link_r.miss = NULL;
    }

    if (1) {
	lxpart_data->link_l.miss = screen_data->link_l.miss;
    }

    if (bits & RESOURCE_LINK) {
	lxpart_data->link_m = screen_data->lock_m;
    } else {
	lxpart_data->link_m.data = NULL;
	lxpart_data->link_m.free = _x1f4_e4_free_data;
	lxpart_data->link_m.link = _x1f4_e4_link_data;
	lxpart_data->link_m.mode = _x1f4_e4_mode_data;
    }

    if (bits & TEXTFLAT_LINK) {
	lxpart_data->link_e = screen_data->lock_e;
    } else {
	lxpart_data->link_e.data = NULL;
	lxpart_data->link_e.line = NULL;
	lxpart_data->link_e.post = NULL;
	lxpart_data->link_e.push = NULL;
    }

    status = link_part(lxpart_data, bits, screen_data);

    return status;
}


static int
line_data(struct lxpart_type *lxpart_data, unsigned bits,
	  const struct screen_type *screen_data, void *data, void *text,
	  void **args)
{
    struct cxpart_type *cxpart_data;
    struct lxtext_type *lxtext_data;
    struct x1f4_linetext_type *linetext_data;
    unsigned miss;

    cxpart_data = screen_data->link_l.cxpart_data;
    miss = screen_data->link_l.miss;

    linetext_data = data;
    lxtext_data = text;

    for (; miss; miss--) {
	line_sail(lxpart_data, screen_data, lxtext_data, cxpart_data);

	if (bits & AUTOLINK_LINK) {
	    _libx1f4i0_lxcall_side_line
		(lxtext_data, bits, cxpart_data, linetext_data, args,
		 sizeof(ever_line) / sizeof(struct line_type), ever_line);
	    linetext_data += sizeof(ever_line) / sizeof(struct line_type);
	}

	lxtext_data++;

	cxpart_data++;
    }

    if (1) {
	_libx1f4i0_lxcall_zero_line(linetext_data);
    }

    return 0;
}


static int
line_fail(struct lxpart_type *lxpart_data, unsigned bits, void *side,
	  void *data)
{
    struct x1f4_operator_type **operator;
    struct x1f4_linetext_type *linetext_data;
    struct tide_type *tide_data;
    unsigned miss;

    miss = lxpart_data->link_l.miss;

    linetext_data = data;

    operator = side;

    tide_data = side;
    tide_data -= miss;

    for (; miss; miss--) {
	if (bits & AUTOLINK_LINK) {
	    *operator = &tide_data->operator;

	    tide_data->operator = side_o[0];
	    tide_data->operator.args = linetext_data->function.args;
	    tide_data->operator.type = linetext_data->function.type;
	    tide_data->operator.extension2 = &tide_data->tidespan;
	    tide_data->tidespan.bits = X1f4_E4_CALL_LINK;
	    tide_data->tidespan.line = linetext_data;

	    linetext_data++;

	    tide_data++;

	    operator++;
	}
    }

    *operator = NULL;

    return 0;
}


static int
line_part(struct lxpart_type *lxpart_data, unsigned bits,
	  const struct screen_type *screen_data, void *text)
{
    struct cxpart_type *cxpart_data;
    struct lxtext_type *lxtext_data;
    struct x1f4_datatype_type *datatype_data;
    unsigned miss;

    cxpart_data = screen_data->link_l.cxpart_data;
    miss = screen_data->link_l.miss;

    lxtext_data = text;

    datatype_data = lxpart_data->link_t.datatype_data;

    for (; miss; miss--) {
	datatype_data->context = lxtext_data;

	datatype_data->flat = _libx1f4i0_lxpart_flat_part;
	datatype_data->lead = _libx1f4i0_lxpart_lead_part;
	datatype_data->line = _libx1f4i0_lxpart_line_part;
	datatype_data->link = _libx1f4i0_lxpart_link_part;
	datatype_data->name = cxpart_data->name;
	datatype_data->shut = _libx1f4i0_lxpart_flat_part;
	datatype_data->slip = _libx1f4i0_lxpart_slip_part;
	datatype_data->size = cxpart_data->size;
	datatype_data->type = cxpart_data->type;

	cxpart_data++;

	lxtext_data++;

	datatype_data++;
    }

    {
	datatype_data->context = NULL;

	datatype_data->flat = NULL;
	datatype_data->lead = NULL;
	datatype_data->line = NULL;
	datatype_data->link = NULL;
	datatype_data->name = NULL;
	datatype_data->shut = NULL;
	datatype_data->slip = NULL;
	datatype_data->size = 0;
	datatype_data->type = 0;
    }

    return 0;
}


static int
line_sail(struct lxpart_type *lxpart_data,
	  const struct screen_type *screen_data,
	  struct lxtext_type *lxtext_data,
	  struct cxpart_type *cxpart_data)
{
    lxtext_data->copy = cxpart_data->copy;
    lxtext_data->flat = cxpart_data->flat;
    lxtext_data->high = cxpart_data->high;
    lxtext_data->line = cxpart_data->line;
    lxtext_data->part = lxpart_data;
    lxtext_data->shut = cxpart_data->shut;
    lxtext_data->text = cxpart_data->context;
    lxtext_data->trap = cxpart_data->trap;

    lxtext_data->nodetype[0].context = lxtext_data;

    _libx1f4i0_lxpart_time_part(lxtext_data->nodetype, cxpart_data);

    return 0;
}


static int
link_part(struct lxpart_type *lxpart_data, unsigned bits,
	  const struct screen_type *screen_data)
{
    int status;
    unsigned args_class, line_class, miss;
    void *data;

    pick_miss(&args_class, &line_class, screen_data);

    miss = screen_data->link_l.miss;

    status = lxpart_data->link_w.link
	(lxpart_data->link_w.data, &data,
	 (args_class * sizeof(int) + sizeof(struct x1f4_datatype_type)
	  + line_class * sizeof(struct x1f4_linetext_type)
	  + sizeof(struct lxtext_type) + sizeof(struct tide_type)
	  + sizeof(struct x1f4_operator_type *))
	 * miss + sizeof(struct x1f4_datatype_type)
	 + sizeof(struct x1f4_linetext_type)
	 + sizeof(struct x1f4_operator_type *));
    if (status) {
	status = LINK_ERROR;
    } else {
	void *args, *flat, *part, *side;

	part = (struct x1f4_linetext_type *) data + miss * line_class + 1;
	flat = (struct lxtext_type *) part + miss;
	side = (struct x1f4_datatype_type *) flat + miss + 1;
	side = (struct tide_type *) side + miss;
	args = (struct x1f4_operator_type **) side + miss + 1;

	lxpart_data->link_f.data =
	    (struct x1f4_linetext_type *) data + miss * line_class;
	lxpart_data->link_f.part = part;
	lxpart_data->link_f.side = side;
	lxpart_data->link_f.text = data;

	lxpart_data->link_t.datatype_data = flat;

	line_data(lxpart_data, bits, screen_data, data, part, &args);
	line_fail(lxpart_data, bits, side, data);

	line_part(lxpart_data, bits, screen_data, part);
    }

    return status;
}


static int
pick_miss(unsigned *args, unsigned *line,
	  const struct screen_type *screen_data)
{
    unsigned args_count = 0, line_count = 0;

    args_count += ever_line_count;
    line_count += sizeof(ever_line) / sizeof(struct line_type);

    *args = args_count;
    *line = line_count;

    return 0;
}


static void
line_sset(u_case_args_____1)
{
    int *line;

    line = *args;

    linetext_data->function.args = line;

    *line++ = cxflat(screen)->type;
    *line++ = cxflat(screen)->type;
    *line++ = X1f4_E4_POST_FLAT | X1f4_E4_POST_XSET;
    *line++ = 0;

    *args = line;

    linetext_data->function.type = cxflat(screen)->type;
}


int
x1f4_init_lxpart(void **lxpart, unsigned bits,
		 const struct screen_type *screen_data)
{
    int (*link) (void *, void **, unsigned), status;
    void *data, *text;

    if (bits & CODELINK_LINK) {
	link = screen_data->lock_w.link;
	text = screen_data->lock_w.data;
    } else {
	link = _x1f4_e4_link_data;
	text = (void *) 0;
    }

    status = link(text, &data, sizeof(struct lxpart_type));
    if (status) {
	status = LINK_ERROR;
    } else {
	status = init_part(data, bits, screen_data);
	if (status) {
	    if (bits & CODELINK_LINK) {
		screen_data->lock_w.free(text, data);
	    } else {
		_x1f4_e4_free_data(text, data);
	    }
	} else {
	    *lxpart = data;
	}
    }

    return status;
}
