/*
 * lxpart-a.b.c
 * Copyright (C) 2008-2011, 2013-2015, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>

#include <exerrors.h>
#include <lxfset.h>
#include <lxlead.h>
#include <lxpart-inter.h>
#include <lxpart-types.h>
#include <nl.h>
#include <tcvset.h>

#define tccase(vset) \
    ((struct x1f4_tccase_type *) (vset))

int
x1f4_copy_lxpart(void *subtext, void *bppp, void **copy, const void *part,
		 const struct x1f4_nodelink_type *nodelink_data)
{
    int status;
    struct lxpart_type *lxpart_data;
    struct lxtext_type *lxtext_data;
    void *dxpart;

    lxtext_data = subtext;

    lxpart_data = lxtext_data->part;

    status = lxpart_data->link_m.link
	(lxpart_data->link_m.data, &dxpart,
	 __struct_dxpart_type_alloc_size__ + lxtext_data->trap);
    if (status) {
	status = _libx1f4i0_lxpart_stat_link(lxpart_data);
    } else {
	void *bqpp;

	do {
	    void *record;

	    if (bppp) {
		bqpp = bppp;
	    } else {
		status = x1f4_init_lxfset(&bqpp);
		if (status) {
		    status = _libx1f4i0_lxpart_stat_link(lxpart_data);
		    if (1) {
			break;
		    }
		}
	    }

	    status = x1f4_post_lxfset(bqpp, &part, &record);
	    if (status) {
		status = _libx1f4i0_lxpart_stat_link(lxpart_data);
		if (1) {
		    break;
		}
	    } else {
		void *pset;
		const void *text;

		dxpart = pick_dxpart(dxpart);

		{
		    {
			void **node;

			node = record;
			node[0] = (void *) part;
			node[1] = dxpart;
		    }
		}

		lock_excase(dxpart)->nodetype_data =
		    lxtext_data->nodetype;

		lock_excase(dxpart)->slip = NULL;

		lock_expart(dxpart)->call = 1;

		_libx1f4i0_init_tcvset(&pset, &text);

		status = tccase(text)->push
		    (&pset, &text, nodelink_data, &lxpart_data->link_m);
		if (status) {
		    status = X1f4_EX_CRITICAL;

		    _libx1f4i0_fini_tcvset(&pset, &text, &lxpart_data->link_m);
		} else {
		    struct x1f4_caselink_type *caselink_data = NULL;

		    lock_expart(dxpart)->pset = pset;
		    lock_expart(dxpart)->text = text;

		    status = lxtext_data->copy
			(bqpp, lxtext_data->text, dxpart, part,
			 &caselink_data);
		    if (status) {
			x1f4_lose_lxlead(caselink_data, &lxpart_data->link_m);

			_libx1f4i0_fini_tcvset
			    (&pset, &text, (void *) &lxpart_data->link_m);
		    } else {
			status = x1f4_beta_lxlead
			    (bqpp, caselink_data, &lxpart_data->link_m);
			if (status) {
			    _libx1f4i0_lxpart_side_part(lxpart_data, dxpart);
			} else {
			    *copy = dxpart;
			}

			dxpart = NULL;
		    }
		}
	    }
	} while (0);

	if (bqpp) {
	    if (bppp) {
	    } else {
		int excess;

		excess = x1f4_fini_lxfset(&bqpp);
		if (excess) {
		    status = _libx1f4i0_lxpart_stat_free(lxpart_data, status);
		}
	    }
	}

	if (dxpart) {
	    lxpart_data->link_m.free
		(lxpart_data->link_m.data, lock_cxpart(dxpart));
	}
    }

    return status;
}
