/*
 * qscc.9.c
 * Copyright (C) 2016, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <qscc-names.h>
#include <qscc-types.h>

#define qsrate(qsrate)			((struct qsrate_type *) (qsrate))

static int half_node(struct fpnode_type *, unsigned, unsigned,
		     struct qsrate_type *, unsigned, unsigned *);
static int long_node(struct fpnode_type *, unsigned, unsigned,
		     struct qsrate_type *, unsigned, unsigned *);
static int tall_rate(struct qsrate_type *, unsigned);

static int
half_node(struct fpnode_type *fpnode_data, unsigned half, unsigned rate,
	  struct qsrate_type *qsrate_data, unsigned sale, unsigned *rule)
{
    int delete = 0;
    unsigned i;

    i = half >> 1;
    for (; i; i--) {
	if (fpnode_data->node) {
	    fpnode_data += 2;
	} else {
	    delete = 12;
	    if (1) {
		break;
	    }
	}
    }

    if (delete) {
    } else {
	delete =
	    long_node(fpnode_data - half, half, rate, qsrate_data, sale, rule);
    }

    return delete;
}


static int
long_node(struct fpnode_type *fpnode_data, unsigned half, unsigned rate,
	  struct qsrate_type *qsrate_data, unsigned sale, unsigned *rule)
{
    int delete;

    half >>= 1;
    if (half) {
	struct fpnode_type *fpnode_text;
	unsigned call;

	fpnode_text = fpnode_data + half;

	call = fpnode_text->call;
	if (sale < call) {
	    delete = 3;
	} else {
	    delete = long_node
		(fpnode_data, half, rate, qsrate_data, call, rule);
	    if (delete) {
	    } else {
		delete = long_node
		    (fpnode_text, half, rate, qsrate_data, sale - call, rule);
	    }
	}
    } else {
	if (fpnode_data->node) {
	    if (rate) {
		unsigned node = 0;

		delete = half_node
		    (fpnode_data->star, 1 << qsrate_data->link_a.fplist.lock,
		     rate - 1, qsrate_data, sale, &node);
		if (delete) {
		} else {
		    if (node ^ fpnode_data->node) {
			delete = 5;
		    } else {
			if (node < 1 << qsrate_data->link_a.fplist.lock >> 1) {
			    delete = 7;
			} else {
			    (*rule)++;
			}
		    }
		}
	    } else {
		if (sale ^ fpnode_data->node) {
		    delete = 4;
		} else {
		    if (sale < qsrate_data->link_a.fplist.line >> 1) {
			delete = 6;
		    } else {
			delete = 0;

			(*rule)++;
		    }
		}
	    }
	} else {
	    delete = 0;
	}
    }

    return delete;
}


static int
tall_rate(struct qsrate_type *qsrate_data, unsigned rate)
{
    int delete;

    if (qsrate_data->link_e.fplink_text) {
	struct fplink_type *fplink_data;

	fplink_data = qsrate_data->link_e.fplink_data;
	while (fplink_data) {
	    fplink_data = fplink_data->fplink_data;
	    rate--;
	}

	delete = rate ? 11 : 0;
    } else {
	delete = 10;
    }

    return delete;
}


int
x1f4_flow_qsrate(void *qsrate)
{
    int delete;
    unsigned rate, sale;

    sale = qsrate(qsrate)->link_a.fplist.size;

    rate = qsrate(qsrate)->link_a.fplist.rate;
    if (rate) {
	unsigned node = 0;

	rate--;

	delete = long_node
	    (qsrate(qsrate)->link_a.fplist.fpnode,
	     1 << qsrate(qsrate)->link_a.fplist.lock, rate, qsrate, sale,
	     &node);
	if (delete) {
	} else {
	    if (node ^ qsrate(qsrate)->link_a.fplist.node) {
		delete = 1;
	    } else {
		delete = tall_rate(qsrate, rate);
	    }
	}
    } else {
	if (3 * qsrate(qsrate)->link_a.fplist.line >> 1 < sale) {
	    delete = 13;
	} else {
	    if (qsrate(qsrate)->link_e.fplink_data) {
		delete = 8;
	    } else {
		if (qsrate(qsrate)->link_e.fplink_text) {
		    delete = 9;
		} else {
		    delete = 0;
		}
	    }
	}
    }

    return delete;
}
