/*
 * s5xv.6.c
 * Copyright (C) 2009-2010, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>
#include <string.h>

#include <s5xx-defs.h>
#include <s5xx-names.h>
#include <s5xx-types.h>

#define fast(miss)			((const byte *) (miss))

#define s5deck(s5deck)			((struct s5deck_type *) (s5deck))

#define fpnode_data(fpnode) \
    ((struct fpnode_type **) ((struct fpnode_type *) (fpnode) + 1))

static int pass_name(void *, const byte *, unsigned, struct fpnode_type *);

static int
pass_name(void *s5deck, const byte *name, unsigned size,
	  struct fpnode_type *fpnode_data)
{
    int status;
    struct fpnode_type **fpnode, **fqnode = NULL, *fqnode_data = NULL;
    unsigned call, rate, tier = 0;

    fpnode = &s5deck(s5deck)->link_a.fpdeck.fpnode_data;

    size <<= 8;

    call = fpnode_data->data.call;
    rate = call & 255;
    while (rate) {
	fqnode_data = fpnode_data;

	fqnode = fpnode;

	if (call < size) {
	    tier = (1 + (rate | name[call >> 8])) >> 8;
	} else {
	    tier = 0;
	}

	fpnode = fpnode_data(fpnode_data) + tier;

	fpnode_data = *fpnode;

	call = fpnode_data->data.call;
	if (1) {
	    rate = call & 255;
	}
    }

    if (call ^ size) {
	status = 0;
    } else {
	if (memcmp
	    (*(const char **) ((void **) (fpnode_data + 1) + 1), name,
	     size >> 8)) {
	    status = 0;
	} else {
	    status = s5deck(s5deck)->link_m.free
		(s5deck(s5deck)->link_m.data, fpnode_data);
	    if (status) {
		status = FREE_ERROR;
	    } else {
		if (fqnode) {
		    fpnode_data = fpnode_data(fqnode_data)[1 - tier];

		    status = s5deck(s5deck)->link_m.free
			(s5deck(s5deck)->link_m.data, fqnode_data);
		    if (status) {
			status = FREE_ERROR;
		    } else {
		    }

		    /*
		     * edit tree even for free failure
		     */

		    *fqnode = fpnode_data;
		} else {
		    s5deck(s5deck)->link_a.fpdeck.fpnode_data = NULL;
		}
	    }
	}
    }

    return status;
}


int
x1f4_pass_s5path(void *s5deck, const char *name, unsigned size)
{
    int status;
    struct fpnode_type *fpnode_data;

    fpnode_data = s5deck(s5deck)->link_a.fpdeck.fpnode_data;
    if (fpnode_data) {
	status = pass_name(s5deck, fast(name), size, fpnode_data);
    } else {
	status = 0;
    }

    return status;
}
