/*
 * tc10.c
 * Copyright (C) 2002-2010, 2013, 2016, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <aime.h>
#include <inter.h>
#include <sf.h>
#include <tfxx.h>

extern const struct x1f4_function_type _libx1f4i0_e4_io[];

static int pick_name(const char *, unsigned, const void *,
		     const struct x1f4_function_type **);
static int test_cast(void *, const char *, unsigned);

static void usage(void);

static int
pick_name(const char *name, unsigned size, const void *text,
	  const struct x1f4_function_type **function)
{
    int status;

    status = libx1f4i0_select_function(name, size, x1f4_e4_defaults, function);
    if (status) {
	status = libx1f4i0_select_function
	    (name, size, _libx1f4i0_e4_io, function);
    }

    return status;
}


static int
test_cast(void *cast, const char *name, unsigned size)
{
    return 1 ^ fwrite(name, size, 1, cast);
}


static void
usage(void)
{
    puts("Usage: tc10 [OPTIONS] PROGRAM\n\
Execute PROGRAM.\n\
\n\
Options:\n\
  -c                            execute the PROGRAM program instead program\n\
				read from the PROGRAM file\n\
      --help			display this help and exit\n\
      --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int immediate = 0, status = 0;

    unsetenv("POSIXLY_CORRECT");

    {
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "help",	    0x00,   NULL,   0x68    },
		{   "version",	    0x00,   NULL,   0x76    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "c", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'c':
		immediate = 1;
		break;
	    case 'h':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}
    }

    if (argc - optind != 1) {
	fprintf(stderr, "%s: wrong number of arguments\nTry `%s --help' for "
		"more information.\n", argv[0], argv[0]);

	return 1;
    } else {
	unsigned size;
	void *data;

	if (immediate) {
	    data = argv[optind];
	} else {
	    status = libx1f4i0_read_file(&data, &size, argv[optind], 1);
	    if (status) {
		libx1f4i0_stat_failure(argv[0], status, argv[optind]);
	    }
	}

	if (status) {
	} else {
	    struct x1f4_c1_type c1;
	    struct x1f4_c1record_type c1record;
	    unsigned flags = X1f4_C1_BCOLLECT | X1f4_C1_DDACCESS;
	    void *program;

	    if (immediate) {
	    } else {
		char *mine;

		mine = data;
		mine[size] = 0;
	    }

	    c1.function_set.get = pick_name;
	    x1f4_llink_operator1s(&c1.operator1s);
	    x1f4_llink_operator2s(&c1.operator2s);
	    c1.variable_set.context = NULL;

	    set_xxrecord(c1, &c1record);

	    status = x1f4_init_program(&program, data, flags, &c1);
	    if (status) {
		if (status == X1f4_C1_ALLOC_ERROR) {
		    perror(argv[0]);
		} else {
		    fprintf(stderr, "%s: cannot parse `%s'\n", argv[0],
			    argv[optind]);
		    fprintf(stderr, "%s: ", argv[0]);
		    x1f4_stat_program
			(stderr, test_cast, data, &c1record, NULL);
		    fprintf(stderr, "\n");
		}
	    } else {
		status = x1f4_link_program(program);
		if (status) {
		    fprintf(stderr, "%s: cannot execute `%s'\n", argv[0],
			    argv[optind]);
		}

		x1f4_fini_program(&program);
	    }

	    if (immediate) {
	    } else {
		free(data);
	    }
	}
    }

    return status;
}
