/*
 * tf22.c
 * Copyright (C) 2002-2016, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <getopt.h>
#include <signal.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/wait.h>

#include <aime.h>
#if !defined HAVE_LIBx1f4i0
# include <cardinal-wx.h>
# include <integral.v-d.h>
#endif				/* !HAVE_LIBx1f4i0 */
#include <inter.h>

#define FILE_BITS \
    X1f4_LXFILE_SETCLOSE

#define PORT_BITS \
    X1f4_LXPORT_SETCLOSE

#define l_MODE(e, output) \
    {									      \
	X1f4_E4_C_MODE *l;						      \
									      \
	l = (void *) (output);						      \
	*l = (e);							      \
    }

#define context(___c) \
    ((struct context_type *) (___c))

#define e4less(___c) \
    (context(___c)->roll[context(___c)->times << 1])

typedef struct context_type {
    char **argv;
    const char *program, *self;
    int argc, class, error, immediate,
	(*line)(struct context_type *, unsigned *), sequenced;
    struct x1f4_e2line_type e2line;
    struct x1f4_indexset_type *indexset_data;
    unsigned call, times;
    void *logic, **roll;
    const void *data;
} context_type;

typedef struct library_type {
    int (*free)(void *, void *), sequence;
    struct x1f4_linetext_type functions[2];
    void *data;
} library_type;

static int copy_error(void *, const char *, unsigned);
static int copy_output(void *, const char *, unsigned);
static int ever_list(void *, void *, struct x1f4_variable_type *);
static int fini_sequence(void **);
static int fini_xset(struct context_type *);
static int flat_text(void *);
static int flush_error(struct context_type *, const char *);
static int free_call(void *, void *);
static int fset_sequence(void *, const struct x1f4_linetext_type **);
static int init_xset(struct context_type *);
static int lead_less(void **, struct context_type *);
static int line_flat(void *, int);
static int line_program(struct context_type *, unsigned *);
static int line_shuffle(struct context_type *, unsigned *);
static int link_e4lf(void *, void **, unsigned);
static int link_call(void *, void **, unsigned);
static int list_text(void *, const void *);
static int lock_e4ll(void *, const struct x1f4_function_type *);
static int long_fine(struct context_type *, const char *, void **);
static int long_line(struct context_type *, const char *, void **, void *);
static int mode_call(void *, void **, unsigned);
static int mode_sequence(void *, void *, void **);
static int note_sequence(const struct x1f4_lxnote_type **);
static int pick_call(void *, void ***, int (*)(void *), unsigned, void **);
static int pick_e4lf(void *, void ***, int (*)(void *), unsigned, void **);
static int pick_program(struct context_type *, unsigned *);
static int pick_shuffle(struct context_type *, unsigned *);
static int pipe_e4ll(void *, void *);
static int post_flat(void *);
static int push_case(struct context_type *, const char *, const void *);
static int push_e4ll(void *, const struct x1f4_function_type *);
static int push_fine(struct context_type *, const char *, const void *);
static int push_flat(void *, const char *, unsigned);
static int push_line(struct context_type *, const char *, const void *);
static int push_port(struct context_type *, int, char **);
static int roll_fine(struct context_type *, const char *, void *);
static int roll_line(struct context_type *, const char *, void *, void *);
static int seek_line(struct context_type *, unsigned, unsigned *);
static int set_composer(struct context_type *, void **,
			int (**)(void *, int, int,
				 const struct x1f4_linetext_type **));
static int slip_call(void *, void *, void *text, int (*)(void *));
static int test_cast(void *, const char *, unsigned);
static int tile_sequence(void **, unsigned, struct x1f4_lxwide_type *,
			 unsigned, const void *);
static int wait_some(struct context_type *);

static void SIGCHLD_handler(int);
static void list_0010(struct context_type *);
static void list_0220(struct context_type *);
static void usage(void);

static int *static_class;
static const int c_____m__[] = {
/* *INDENT-OFF* */
    X1f4_E4_MODE
/* *INDENT-ON* */
};
static sig_atomic_t SIGCHLD_received;
static struct x1f4_lxnear_type lxnear;
static const struct x1f4_lxnote_type sequence_note = {
/* *INDENT-OFF* */
    NULL,
    NULL,
    NULL,
    NULL,
    fini_sequence,
    fset_sequence,
    NULL,
    NULL,
    NULL,
    tile_sequence,
    NULL,
    NULL,
    0
/* *INDENT-ON* */
};
const static struct x1f4_lxtile_type tiles[] = {
/* *INDENT-OFF* */
    {	note_sequence,		0,		NULL		},
    {	x1f4_note_lxcast,	0,		NULL		},
    {	x1f4_note_lxdata,	0,		NULL		},
    {	x1f4_note_lxdate,	0,		NULL		},
    {	x1f4_note_lxdecq,	0,		NULL		},
    {	x1f4_note_lxfile,	FILE_BITS,	NULL		},
    {	x1f4_note_lxline,	0,		NULL		},
    {	x1f4_note_lxlist,	0,		NULL		},
    {	x1f4_note_lxnear,	0,		&lxnear		},
    {	x1f4_note_lxport,	PORT_BITS,	NULL		},
    {	x1f4_note_lxtext,	0,		NULL		},
    {	x1f4_note_lxtrap,	0,		NULL		}
/* *INDENT-ON* */
};
static const struct x1f4_textpipe_type near_pipes[] = {
/* *INDENT-OFF* */
    {	"o",		copy_output,		1,	NULL	},
    {	"v",		copy_error,		1,	NULL	}
/* *INDENT-ON* */
};

static int
copy_error(void *none, const char *data, unsigned size)
{
    fwrite(data, size, 1, stderr);

    return 0;
}


static int
copy_output(void *none, const char *data, unsigned size)
{
    fwrite(data, size, 1, stdout);

    return 0;
}


static int
ever_list(void *none, void *data, struct x1f4_variable_type *variable_data)
{
    puts(variable_data->name);

    return 0;
}


static int
fini_sequence(void **sequence)
{
    int status;
    struct library_type *library_data;

    library_data = *sequence;
    if (library_data->free) {
	status = library_data->free(library_data->data, library_data);
    } else {
	free(library_data);
	status = 0;
    }

    return status;
}


static int
fini_xset(struct context_type *context_data)
{
    return x1f4_flat_indexset(context_data->indexset_data);
}


static int
flat_text(void *context)
{
    int status;
    struct context_type *context_data;
    struct x1f4_indexset_type *indexset_data;

    context_data = context;

    indexset_data = context_data->indexset_data;

    status = indexset_data->autodeck_set.deck
	(&indexset_data->autodeck_set.text);

    if (SIGCHLD_received) {
	int excess;

	excess = wait_some(context_data);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
flush_error(struct context_type *context_data, const char *program)
{
    if (context_data->error) {
    } else {
	fflush(stdout);
	fprintf(stderr, "%s:", context_data->self);
	if (program) {
	    fprintf(stderr, " %s:", program);
	}
	if (1) {
	    int (*line)(struct context_type *, unsigned *);

	    line = context_data->line;
	    if (line) {
		unsigned copy, miss;

		line(context_data, &copy);
		seek_line(context_data, copy, &miss);
		fprintf(stderr, " %u:", miss + 1);
	    }
	}

	fprintf(stderr, " cannot execute program\n");
    }

    return 17;
}


static int
free_call(void *context, void *data)
{
    void *e2list;

    x1f4_fast_e4ll(e4less(context), &e2list);

    return x1f4_free_e2list(e2list, data);
}


static int
fset_sequence(void *sequence, const struct x1f4_linetext_type **linetext)
{
    *linetext = ((struct library_type *) sequence)->functions;

    return 0;
}


static int
init_xset(struct context_type *context_data)
{
    struct x1f4_textport_type textport;
    unsigned textport_bits = 0;

    textport_bits = X1f4_TEXTPORT_TEXTFLAT;

    if (context_data->sequenced) {
	textport_bits |= X1f4_TEXTPORT_AUTOLINK;

	textport.autolink_set.data = context_data;
	textport.autolink_set.free = free_call;
	textport.autolink_set.link = link_call;
	textport.autolink_set.mode = mode_call;
	textport.autolink_set.pick = pick_call;
	textport.autolink_set.slip = slip_call;
    }

    textport.lasttype_set.type = X1f4_E4_LAST;

    textport.tilelong_set.lxtile_data = tiles;
    textport.tilelong_set.miss =
	sizeof(tiles) / sizeof(struct x1f4_lxtile_type);

    lxnear.nearpipe_set.textpipe_data = near_pipes;
    lxnear.nearpipe_set.miss = 2;

    textport.textflat_set.data = context_data;
    textport.textflat_set.line = line_flat;
    textport.textflat_set.post = post_flat;
    textport.textflat_set.push = push_flat;

    textport.autodeck_set.class = &context_data->class;

    return x1f4_fast_indexset
	(context_data->indexset_data, textport_bits, &textport);
}


static int
lead_less(void **less, struct context_type *context_data)
{
    struct x1f4_e2line_type *e2line_data;
    struct x1f4_e4less_type e4line;

    e2line_data = &context_data->e2line;

    e2line_data->bits = X1f4_E2LIST_FLATLINE;

    e2line_data->e2list.flatline_set.class = &context_data->class;

    e4line.misslink_set.data = e2line_data;

    x1f4_type_e2list(&e4line.misslink_set);

    return x1f4_init_e4ll(less, X1f4_E4LL_MISSLINK, &e4line);
}


static int
line_flat(void *context, int class)
{
    struct context_type *context_data;

    context_data = context;

    fflush(stdout);
    fputs(context_data->self, stderr);
    fputs(":", stderr);
    if (context_data->program) {
	fputs(" ", stderr);
	fputs(context_data->program, stderr);
	fputs(":", stderr);
    }
    if (1) {
	int (*line)(struct context_type *, unsigned *);

	line = context_data->line;
	if (line) {
	    unsigned copy, miss;

	    line(context, &copy);
	    seek_line(context, copy, &miss);
	    fprintf(stderr, " %u:", miss + 1);
	}
    }

    fputs(" ", stderr);

    return 0;
}


static int
line_program(struct context_type *context_data, unsigned *copy)
{
    return x1f4_look_routine(context_data->logic, copy);
}


static int
line_shuffle(struct context_type *context_data, unsigned *copy)
{
    return x1f4_look_shuffle(context_data->logic, copy);
}


static int
link_call(void *context, void **link, unsigned size)
{
    void *e2list;

    x1f4_fast_e4ll(e4less(context), &e2list);

    return x1f4_link_e2list(e2list, link, size);
}


static int
link_e4lf(void *context, void **link, unsigned size)
{
    return x1f4_link_e4lf(e4less(context), link, size);
}


static int
list_text(void *context, const void *text)
{
    x1f4_print_htfunction
	(stdout, text, "", ((struct context_type *) context)->indexset_data,
	 x1f4_look_indexset);

    return 0;
}


static int
lock_e4ll(void *context, const struct x1f4_function_type *function_data)
{
    return x1f4_lock_e4ll(e4less(context), function_data);
}


static int
long_fine(struct context_type *context_data, const char *program, void **roll)
{
    int status;
    unsigned times;

    times = context_data->times;

    while (times) {
	unsigned i;

	i = times;
	for (; i; i--) {
	    context_data->logic = *roll;
	    context_data->roll = roll;

	    status = x1f4_step_routine(*roll);
	    if (status == X1f4_EX_COMPLETED) {
		memmove(roll, roll + 1, (i - 1) * sizeof(void *));
		memmove(roll + context_data->times * 2,
			roll + context_data->times * 2 + 1,
			(i - 1) * sizeof(void *));
		times--;

		if (1) {
		    status = 0;
		}
	    } else {
		if (context_data->class) {
		    int excess;
		    void *e2list;

		    context_data->class = 0;

		    x1f4_fast_e4ll(e4less(context_data), &e2list);

		    excess = x1f4_deck_e2list(&e2list);
		    if (excess) {
			if (status) {
			} else {
			    status = excess;
			}
		    }

		    if (SIGCHLD_received) {
			excess = wait_some(context_data);
			if (excess) {
			    if (status) {
			    } else {
				status = excess;
			    }
			}
		    }
		}

		if (status) {
		    flush_error(context_data, program);
		}

		roll++;
	    }

	    if (status) {
		break;
	    }
	}

	if (status) {
	    break;
	}

	roll -= times;
    }

    return status;
}


static int
long_line(struct context_type *context_data, const char *program, void **roll,
	  void *degree)
{
    int status;
    unsigned times;

    times = context_data->times;

    while (times) {
	unsigned i;

	i = times;
	for (; i; i--) {
	    context_data->logic = *roll;
	    context_data->roll = roll;

	    status = x1f4_step_shuffle(*roll);
	    if (status == X1f4_EX_COMPLETED) {
		status = 0;

#if 0
		x1f4_post_shuffle(*roll, degree);
		if (*(X1f4_E4_C_MODE *) degree) {
		    status = *(X1f4_E4_C_MODE *) degree;
		}
#endif				/* 0 */

		memmove(roll, roll + 1, (i - 1) * sizeof(void *));
		memmove(roll + context_data->times * 2,
			roll + context_data->times * 2 + 1,
			(i - 1) * sizeof(void *));
		times--;
	    } else {
		if (context_data->class) {
		    int excess;
		    void *e2list;

		    context_data->class = 0;

		    x1f4_fast_e4ll(e4less(context_data), &e2list);

		    excess = x1f4_deck_e2list(&e2list);
		    if (excess) {
			if (status) {
			} else {
			    status = excess;
			}
		    }

		    if (SIGCHLD_received) {
			excess = wait_some(context_data);
			if (excess) {
			    if (status) {
			    } else {
				status = excess;
			    }
			}
		    }
		}

		if (status) {
		    flush_error(context_data, program);
		}

		roll++;
	    }

	    if (status) {
		break;
	    }
	}

	if (status) {
	    break;
	}

	roll -= times;
    }

    return status;
}


static int
mode_call(void *context, void **link, unsigned size)
{
    void *e2list;

    x1f4_fast_e4ll(e4less(context), &e2list);

    return x1f4_mode_e2list(e2list, link, size);
}


static int
mode_sequence(void *context, void *output, void **input)
{
    struct library_type *library_data;

    library_data = context;

    l_MODE(library_data->sequence, output);

    library_data->sequence++;

    return 0;
}


static int
note_sequence(const struct x1f4_lxnote_type **lxnote)
{
    *lxnote = &sequence_note;

    return 0;
}


static int
pick_call(void *context, void ***text, int (*miss)(void *), unsigned size,
	  void **link)
{
    void *e2list;

    x1f4_fast_e4ll(e4less(context), &e2list);

    return x1f4_pick_e2list(e2list, text, miss, size, link);
}


static int
pick_e4lf(void *context, void ***text, int (*miss)(void *), unsigned size,
          void **link)
{
    return x1f4_pick_e4lf(e4less(context), text, miss, size, link);
}


static int
pick_program(struct context_type *context_data, unsigned *copy)
{
    return x1f4_look_program(context_data->logic, copy);
}


static int
pick_shuffle(struct context_type *context_data, unsigned *copy)
{
    return x1f4_seek_shuffle(context_data->logic, copy);
}


static int
pipe_e4ll(void *context, void *shuffle)
{
    struct x1f4_a1screen_type screen;

    screen.back = lock_e4ll;
    screen.fast = push_e4ll;
    screen.lose = lock_e4ll;

    screen.data = context;

    return x1f4_pipe_shuffle(shuffle, &screen);
}


static int
post_flat(void *context)
{
    ((struct context_type *) context)->error = 1;

    fputs("\n", stderr);

    return 0;
}


static int
push_case(struct context_type *context_data, const char *program,
	  const void *data)
{
    int status;
    unsigned bits = 0;

    if (0) {
    } else {
	bits |= X1f4_A1_SCOMMENT;
    }

    if (x1f4_peer_shuffle(data, bits)) {
	status = push_line(context_data, program, data);
    } else {
	status = push_fine(context_data, program, data);
    }

    return status;
}


static int
push_e4ll(void *context, const struct x1f4_function_type *function_data)
{
    return x1f4_push_e4ll(e4less(context), function_data);
}


static int
push_fine(struct context_type *context_data, const char *program,
	  const void *data)
{
    int status;
    struct x1f4_c1_type c1;
    struct x1f4_c1record_type c1record;
    unsigned flags;
    void *proGram;

    x1f4_xset_program(&c1, &flags, context_data->indexset_data);

    flags |= X1f4_C1_BCOLLECT | X1f4_C1_COMPOSER | X1f4_C1_SIDELIST;

    c1.bcollect_set.c1record_data = &c1record;
    set_composer(context_data, &c1.composer_set.context, &c1.composer_set.get);
    x1f4_pset_lxcast
	((void *) c1.composer_set.context, (void *) &c1.sidetype_set.miss);

    status = x1f4_init_program(&proGram, data, flags, &c1);
    if (status) {
	if (status == X1f4_C1_ALLOC_ERROR) {
	    perror(context_data->self);
	} else {
	    fprintf(stderr, "%s: cannot parse ", context_data->self);
	    if (program) {
		fprintf(stderr, "`%s'", program);
	    } else {
		fprintf(stderr, "program");
	    }
	    fprintf(stderr, "\n");
	    fprintf(stderr, "%s: ", context_data->self);
	    c1record.pick = 1;
	    x1f4_stat_program
		(stderr, test_cast, data, &c1record,
		 &context_data->indexset_data->eelookup_set.eelookup);
	    fprintf(stderr, "\n");
	}
    } else {
	if (0) {
	} else {
	    context_data->class = 0;
	    context_data->error = 0;

	    context_data->logic = proGram;

	    static_class = &context_data->class;

	    if (context_data->sequenced) {
		status = roll_fine(context_data, program, proGram);
	    } else {
		context_data->line = pick_program;

		status = x1f4_note_program
		    (proGram, &context_data->class, context_data, flat_text);
		if (status) {
		    flush_error(context_data, program);
		}
	    }
	}

	x1f4_fini_program(&proGram);
    }

    return status;
}


static int
push_flat(void *context, const char *data, unsigned size)
{
    fwrite(data, size, 1, stderr);

    return 0;
}


static int
push_line(struct context_type *context_data, const char *program,
	  const void *data)
{
    int status;
    struct x1f4_a1_type a1;
    struct x1f4_a1record_type a1record;
    unsigned flags;
    void *proGram;

    x1f4_xset_shuffle(&a1, &flags, context_data->indexset_data);

    flags |= X1f4_A1_BCOLLECT | X1f4_A1_COMPOSER | X1f4_A1_SIDELIST;

    if (context_data->sequenced) {
	a1.autolead_set.context = context_data;
	a1.autolead_set.link = link_e4lf;
	a1.autolead_set.pick = pick_e4lf;
    }

    a1.bcollect_set.a1record_data = &a1record;
    set_composer(context_data, &a1.composer_set.context, &a1.composer_set.get);
    x1f4_pset_lxcast
	((void *) a1.composer_set.context, (void *) &a1.sidetype_set.miss);

    status = x1f4_init_shuffle(&proGram, data, flags, &a1);
    if (status) {
	if (status == X1f4_A1_ALLOC_ERROR) {
	    perror(context_data->self);
	} else {
	    fprintf(stderr, "%s: cannot parse ", context_data->self);
	    if (program) {
		fprintf(stderr, "`%s'", program);
	    } else {
		fprintf(stderr, "program");
	    }
	    fprintf(stderr, "\n");
	    fprintf(stderr, "%s: ", context_data->self);
	    a1record.pick = 1;
	    x1f4_stat_shuffle
		(stderr, test_cast, data, &a1record,
		 &context_data->indexset_data->eelookup_set.eelookup);
	    fprintf(stderr, "\n");
	}
    } else {
	if (context_data->sequenced) {
	    status = pipe_e4ll(context_data, proGram);
	} else {
	    status = x1f4_pipe_e4ll
		(context_data->indexset_data->autodeck_set.less, proGram);
	}
	if (status) {
	    perror(context_data->self);
	} else {
	    X1f4_E4_C_MODE degree;

	    context_data->class = 0;
	    context_data->error = 0;

	    context_data->logic = proGram;

	    static_class = &context_data->class;

	    if (context_data->sequenced) {
		status = roll_line(context_data, program, proGram, &degree);
	    } else {
		context_data->line = pick_shuffle;

		status = x1f4_long_shuffle
		    (proGram, &context_data->class, context_data,
		     flat_text, &degree);
		if (status) {
		    flush_error(context_data, program);
		} else {
		    status = degree;
		}
	    }
	}

	x1f4_fini_shuffle(&proGram);
    }

    return status;
}


static int
push_port(struct context_type *context_data, int argc, char **argv)
{
    int status;
    struct x1f4_indexset_type indexset;

    context_data->indexset_data = &indexset;

    status = init_xset(context_data);
    if (status) {
    } else {
	const char *program;

	program = argv[optind];
	optind++;

	SIGCHLD_received = 0;

	signal(SIGCHLD, SIGCHLD_handler);

	if (context_data->immediate) {
	    if (0) {
	    } else {
		context_data->data = program;

		context_data->program = NULL;

		status = push_case(context_data, NULL, program);
	    }
	} else {
	    unsigned size;
	    void *data;

	    status = libx1f4i0_read_file(&data, &size, program, 1);
	    if (status) {
		libx1f4i0_stat_failure(argv[0], status, program);
	    } else {
		if (1) {
		    char *mine;

		    mine = data;
		    mine[size] = 0;
		}

		context_data->data = data;

		context_data->program = program;

		status = push_case(context_data, program, data);

		free(data);
	    }
	}

	signal(SIGCHLD, SIG_DFL);

	if (1) {
	    int excess;

	    excess = fini_xset(context_data);
	    if (excess) {
		if (status) {
		} else {
		    status = excess;
		}
	    }
	}
    }

    return status;
}


static int
roll_fine(struct context_type *context_data, const char *program,
	  void *proGram)
{
    int status;
    unsigned times;
    void **roll;

    x1f4_rail_program(proGram);

    times = context_data->times;

    context_data->line = line_program;

    roll = malloc(sizeof(void *) * times << 2);
    if (!roll) {
	status = -1;
    } else {
	int excess;
	unsigned i;

	roll += times * 2;

	i = times;
	for (; i; i--) {
	    status = lead_less(roll, context_data);
	    if (status) {
		break;
	    } else {
		roll++;
	    }
	}

	if (status) {
	} else {
	    memcpy(roll, roll - times, sizeof(void *) * times);

	    roll -= times * 3;

	    i = times;
	    for (; i; i--) {
		status = x1f4_near_routine(roll, proGram);
		if (status) {
		    break;
		} else {
		    x1f4_head_routine(*roll);
		    roll++;
		}
	    }

	    if (status) {
	    } else {
		memcpy(roll, roll - times, sizeof(void *) * times);
		status = long_fine(context_data, program, roll);
	    }

	    i = times - i;
	    for (; i; i--) {
		roll--;
		excess = x1f4_side_routine(roll, proGram);
		if (excess) {
		    if (status) {
		    } else {
			status = excess;
		    }
		}
	    }

	    roll += times * 3;
	}

	i = times - i;
	for (; i; i--) {
	    roll--;
	    excess = x1f4_fini_e4ll(roll);
	    if (excess) {
		if (status) {
		} else {
		    status = excess;
		}
	    }
	}

	roll -= times * 2;

	free(roll);
    }

    return status;
}


static int
roll_line(struct context_type *context_data, const char *program,
	  void *proGram, void *degree)
{
    int status;
    unsigned times;
    void **roll;

    x1f4_rail_shuffle(proGram);

    times = context_data->times;

    context_data->line = line_shuffle;

    roll = malloc(sizeof(void *) * times << 2);
    if (!roll) {
	status = -1;
    } else {
	int excess;
	unsigned i;

	roll += times * 2;

	i = times;
	for (; i; i--) {
	    status = lead_less(roll, context_data);
	    if (status) {
		break;
	    } else {
		roll++;
	    }
	}

	if (status) {
	} else {
	    memcpy(roll, roll - times, sizeof(void *) * times);

	    roll -= times * 3;

	    i = times;
	    for (; i; i--) {
		context_data->roll = roll;

		status = x1f4_near_shuffle(roll, proGram);
		if (status) {
		    break;
		} else {
		    x1f4_head_shuffle(*roll);
		    roll++;
		}
	    }

	    if (status) {
	    } else {
		memcpy(roll, roll - times, sizeof(void *) * times);

		status = long_line(context_data, program, roll, degree);
	    }

	    i = times - i;
	    for (; i; i--) {
		roll--;

		context_data->roll = roll;

		excess = x1f4_side_shuffle(roll, proGram);
		if (excess) {
		    if (status) {
		    } else {
			status = excess;
		    }
		}
	    }

	    roll += times * 3;
	}

	i = times - i;
	for (; i; i--) {
	    roll--;
	    excess = x1f4_fini_e4ll(roll);
	    if (excess) {
		if (status) {
		} else {
		    status = excess;
		}
	    }
	}

	roll -= times * 2;

	free(roll);
    }

    return status;
}


static int
seek_line(struct context_type *context_data, unsigned copy, unsigned *miss)
{
    const char *data;
    unsigned line = 0;

    data = context_data->data;
    while (copy) {
	if (*data == 10) {
	    line++;
	}

	copy--;
	data++;
    }

    *miss = line;

    return 0;
}


static int
set_composer(struct context_type *context_data, void **text,
	     int (**lock)(void *, int, int,
			  const struct x1f4_linetext_type **))
{
    struct x1f4_lxslip_type *lxslip_data;

    lxslip_data = context_data->indexset_data->sliplong_set.lxslip_data;
    while (lxslip_data->note != x1f4_note_lxcast) {
	lxslip_data++;
    }

    *lock = x1f4_lock_lxcast;

    *text = lxslip_data->slip;

    return 0;
}


static int
slip_call(void *context, void *data, void *text, int (*miss)(void *))
{
    void *e2list;

    x1f4_fast_e4ll(e4less(context), &e2list);

    return x1f4_slip_e2list(e2list, data, text, miss);
}


static int
test_cast(void *cast, const char *name, unsigned size)
{
    return size ^ fwrite(name, 1, size, cast);
}


static int
tile_sequence(void **sequence, unsigned wide_bits,
	      struct x1f4_lxwide_type *lxwide_data, unsigned bits,
	      const void *data)
{
    int status;
    struct library_type *library_data;

    if (wide_bits & X1f4_LXWIDE_CODELINK) {
	void *library;

	status = lxwide_data->codelink_set.link
	    (lxwide_data->codelink_set.data, &library,
	     sizeof(struct library_type));
	if (status) {
	} else {
	    library_data = library;
	    library_data->data = lxwide_data->codelink_set.data;
	    library_data->free = lxwide_data->codelink_set.free;
	}
    } else {
	library_data = malloc(sizeof(struct library_type));
	if (library_data) {
	    status = 0;
	    library_data->free = NULL;
	} else {
	    status = -1;
	}
    }
    if (status) {
    } else {
	library_data->sequence = 0;

	library_data->functions[0].context = library_data;
	library_data->functions[0].function.args = NULL;
	library_data->functions[0].function.count = 0;
	library_data->functions[0].function.flags =
	    X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK;
	library_data->functions[0].function.function = mode_sequence;
	library_data->functions[0].function.length = 8;
	library_data->functions[0].function.name = "sequence";
	library_data->functions[0].function.type = X1f4_E4_MODE;

	library_data->functions[1].function.name = NULL;

	*sequence = library_data;
    }

    return status;
}


static int
wait_some(struct context_type *context_data)
{
    int status;
    struct x1f4_lxslip_type *lxslip_data;
    void *lxport;

    SIGCHLD_received = 0;

    lxslip_data = context_data->indexset_data->sliplong_set.lxslip_data;
    while (lxslip_data->note != x1f4_note_lxport) {
	lxslip_data++;
    }

    lxport = lxslip_data->slip;

    while (1) {
	pid_t wait;

	wait = waitpid(-1, NULL, WNOHANG);
	if (wait == -1) {
	    status = 0;
	    if (1) {
		break;
	    }
	} else {
	    if (wait) {
		x1f4_wait_lxport(lxport, wait);
	    } else {
		status = 0;
		if (1) {
		    break;
		}
	    }
	}
    }

    return status;
}


static void
SIGCHLD_handler(int signal)
{
    SIGCHLD_received = 1;
    *static_class = 1;
}


static void
list_0010(struct context_type *context_data)
{
    struct x1f4_indexset_type indexset;

    context_data->indexset_data = &indexset;

    context_data->argv = (void *) context_data;

    if (init_xset(context_data)) {
    } else {
	x1f4_lime_mxdeck(indexset.function_set.text, context_data, list_text);

	fini_xset(context_data);
    }
}


static void
list_0220(struct context_type *context_data)
{
    struct x1f4_indexset_type indexset;

    context_data->indexset_data = &indexset;

    context_data->argv = (void *) context_data;

    if (init_xset(context_data)) {
    } else {
	x1f4_list_state(indexset.variable_set.text, NULL, ever_list);

	fini_xset(context_data);
    }
}


static void
usage(void)
{
    puts("Usage: tf22 [OPTIONS] PROGRAM\n\
Execute PROGRAM.\n\
\n\
Options:\n\
  -c				execute the PROGRAM program instead program\n\
				read from the PROGRAM file\n\
  -f, --framed			run function calls in a hierarchical fashion\n\
  -s, --sequenced		flatten call hierarchy\n\
  -x, --parallel N		run PROGRAM parallel N times (-s only)\n\
      --data			list defined constants and exit\n\
      --list			list available functions and exit\n\
      --help			display this help and exit\n\
      --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int list_functions = 0, status = 0;
    struct context_type context;

    context.immediate = 0;

    context.sequenced = 0;
    context.times = 1;

    unsetenv("POSIXLY_CORRECT");

    {
	const char *times_specs = NULL;
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "data",	    0x00,   NULL,   0x02    },
		{   "framed",	    0x00,   NULL,   0x66    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   "list",	    0x00,   NULL,   0x01    },
		{   "parallel",	    0x01,   NULL,   0x78    },
		{   "print",	    0x00,   NULL,   0x50    },
		{   "sequenced",    0x00,   NULL,   0x73    },
		{   "version",	    0x00,   NULL,   0x76    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "fcsx:", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 001:
		list_functions = 1;
		break;
	    case 002:
		list_functions = 2;
		break;
	    case 'c':
		context.immediate = 1;
		break;
	    case 'f':
		context.sequenced = 0;
		break;
	    case 'h':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case 's':
		context.sequenced = 1;
		break;
            case 'x':
                times_specs = optarg;
                break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}

	if (times_specs) {
	    if (x1f4_parse_wxcardinal
		(&context.times, times_specs, NULL, 0)) {
		fprintf(stderr, "%s: cannot parse parallel count specificati"
			"on: `%s'\n", argv[0], times_specs);

		return 1;
	    }
	}
    }

    context.argc = argc - optind;
    context.argv = argv + optind;

    if (list_functions) {
	if (list_functions == 1) {
	    list_0010(&context);
	} else {
	    if (list_functions == 2) {
		list_0220(&context);
	    }
	}
    } else {
	if (context.argv ? !(argc - optind) : (argc - optind) % 3 != 1) {
	    fprintf(stderr, "%s: wrong number of arguments\nTry `%s --help' "
		    "for more information.\n", argv[0], argv[0]);

	    return 1;
	} else {
	    context.self = argv[0];

	    status = push_port(&context, argc, argv);
	}
    }

    return status;
}
