/*
 * tf26.c
 * Copyright (C) 2002-2016, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * MISSING
 *
 * the fiction_type.express_data list not cleared when destroying fiction_type
 * objects
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <getopt.h>
#if defined HAVE_LIBx1f4l0
# include <libx1f4l0.h>
#endif				/* HAVE_LIBx1f4l0 */
#include <signal.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/wait.h>

#include <aime.h>
#if !defined HAVE_LIBx1f4i0
# include <cardinal-wx.h>
# include <integral.v-d.h>
#endif				/* !HAVE_LIBx1f4i0 */
#include <inter.h>
#include <types.h>

#define __ROLL_VALIST__                 1

#if defined HAVE_LIBx1f4l0
#else
# undef __ROLL_VALIST__
# define __ROLL_VALIST__                0
#endif				/* HAVE_LIBx1f4l0 */

#define FILE_BITS \
    X1f4_LXFILE_SETCLOSE

#define NEAR_BITS \
    X1f4_LXNEAR_LONGPIPE

#define PORT_BITS \
    X1f4_LXPORT_SETCLOSE

#define I_TEXT(i)			(*((X1f4_E4_C_TEXT *) (i)))

#define l_MODE(e, output) \
    {									      \
	X1f4_E4_C_MODE *l;						      \
									      \
	l = (void *) (output);						      \
	*l = (e);							      \
    }

#define l_USER(e, output) \
    {									      \
	X1f4_E4_C_USER *l;						      \
									      \
	l = (void *) (output);						      \
	*l = (e);							      \
    }

#define true(e)				1

#define context(___c) \
    ((struct context_type *) (___c))

#define e4less(___c) \
    (((struct fiction_type *) context(___c)->fiction)->turn)

#define VIRTUAL \
    "integer main(void) { virtual26(); return 0; }"

typedef struct express_type {
    int (*call)(void *, int, void *, struct x1f4_dispatch_type **);
    struct express_type *express_data;
    struct x1f4_dispatch_type dispatch;
    void *back, *program;
} express_type;

typedef struct fiction_type {
    struct express_type *express_data;
    struct fiction_type *last_fiction, *next_fiction;
    struct x1f4_fxdata_type record;
    void *device, *input[3], *part, *turn;
} fiction_type;

typedef struct context_type {
    const char *program, *self;
    int class, error, fork, immediate,
	(*line)(struct context_type *, unsigned *), sequenced, through;
    struct fiction_type *fiction_data;
#if __ROLL_VALIST__
    struct list_type dlist;
#endif				/* __ROLL_VALIST__ */
    struct x1f4_e2line_type e2line;
    struct x1f4_function_type *function_data;
    struct x1f4_fxdata_type record;
    struct x1f4_indexset_type *indexset_data;
    struct x1f4_linetext_type linetext[3], positive[1];
    const struct x1f4_linetext_type *long_call, *long_fall, *long_list;
    unsigned call;
    void *device, *fiction, *logic, *pick[4];
#if __ROLL_VALIST__
    void *valist;
#endif				/* __ROLL_VALIST__ */
    void *virtual;
    const void *data;
} context_type;

static int copy_error(void *, const char *, unsigned);
static int copy_output(void *, const char *, unsigned);
static int draw_call(void *, int, void *, struct x1f4_dispatch_type **);
static int ever_list(void *, void *, struct x1f4_variable_type *);
static int fini_valist(struct context_type *);
static int fini_xset(struct context_type *);
static int flat_text(void *);
static int flush_error(struct context_type *, const char *);
static int free_call(void *, void *);
static int init_valist(struct context_type *);
static int init_xset(struct context_type *);
static int land_fine(void *, void *, void **);
static int land_line(void *, void *, void **);
static int land_roll(struct context_type *);
static int lead_less(void **, struct context_type *);
static int line_flat(void *, int);
static int line_program(struct context_type *, unsigned *);
static int line_shuffle(struct context_type *, unsigned *);
static int link_call(void *, void **, unsigned);
static int link_e4lf(void *, void **, unsigned);
static int list_text(void *, const void *);
static int lock_e4ll(void *, const struct x1f4_function_type *);
static int long_fine(struct context_type *, const char *);
static int long_line(struct context_type *, const char *, void *);
static int mode_call(void *, void **, unsigned);
static int part_line(struct context_type *, unsigned, struct x1f4_a1_type *);
static int past_fork(void *, void *, void **);
static int past_line(void *, void *, void **);
static int pick_call(void *, void ***, int (*)(void *), unsigned, void **);
static int pick_e4lf(void *, void ***, int (*)(void *), unsigned, void **);
static int pick_line(void *, void *, void **);
static int pick_program(struct context_type *, unsigned *);
static int pick_shuffle(struct context_type *, unsigned *);
static int pipe_e4ll(void *, void *);
static int posh_line(struct context_type *, unsigned *, struct x1f4_a1_type *,
		     struct x1f4_a1record_type *);
static int post_flat(void *);
static int push_case(struct context_type *, const char *, const void *);
static int push_e4ll(void *, const struct x1f4_function_type *);
static int push_fine(struct context_type *, const char *, const void *);
static int push_flat(void *, const char *, unsigned);
static int push_line(struct context_type *, const char *, const void *);
static int push_port(struct context_type *, int, char **);
static int roll_fine(struct context_type *, const char *, void *);
static int roll_fork(struct context_type *);
static int roll_line(struct context_type *, const char *, void *, void *);
static int rule_fset(struct context_type *, const struct x1f4_linetext_type *);
static int rule_text(struct context_type *);
static int seek_line(struct context_type *, unsigned, unsigned *);
static int set_composer(struct context_type *, void **,
			int (**)(void *, int, int,
				 const struct x1f4_linetext_type **));
static int slip_call(void *, void *, void *text, int (*)(void *));
static int slip_line(struct context_type *, struct fiction_type *);
static int test_cast(void *, const char *, unsigned);
static int wait_some(struct context_type *);

static void SIGCHLD_handler(int);
static void list_0010(struct context_type *);
static void list_0220(struct context_type *);
static void usage(void);

static int *static_class, used_object = X1f4_E4_OBJECT, used_text =
    X1f4_E4_TEXT;
static sig_atomic_t SIGCHLD_received;
static struct x1f4_datapipe_type long_pipes[] = {
/* *INDENT-OFF* */
    {	"b",			x1f4_pipe_lxdata,
	0,			1,		NULL		},
    {	"f",			x1f4_pipe_lxfile,
	0,			1,		NULL		}
/* *INDENT-ON* */
};
static struct x1f4_lxnear_type lxnear;
const static struct x1f4_lxtile_type tiles[] = {
/* *INDENT-OFF* */
    {	x1f4_note_lxcase,	0,		NULL		},
    {	x1f4_note_lxcast,	0,		NULL		},
    {	x1f4_note_lxdash,	0,		NULL		},
    {	x1f4_note_lxdata,	0,		NULL		},
    {	x1f4_note_lxdate,	0,		NULL		},
    {	x1f4_note_lxdecq,	0,		NULL		},
    {	x1f4_note_lxfile,	FILE_BITS,	NULL		},
    {	x1f4_note_lxline,	0,		NULL		},
    {	x1f4_note_lxlist,	0,		NULL		},
    {	x1f4_note_lxnear,	NEAR_BITS,	&lxnear		},
    {	x1f4_note_lxport,	PORT_BITS,	NULL		},
    {	x1f4_note_lxtext,	0,		NULL		},
    {	x1f4_note_lxtrap,	0,		NULL		}
/* *INDENT-ON* */
};
static const struct x1f4_textpipe_type near_pipes[] = {
/* *INDENT-OFF* */
    {	"o",		copy_output,		1,	NULL	},
    {	"v",		copy_error,		1,	NULL	}
/* *INDENT-ON* */
};

static int
back_roll(struct context_type *context_data)
{
    return x1f4_pass_mxpath
	(context_data->indexset_data->function_set.text, "pall", 4);
}


static int
copy_error(void *none, const char *data, unsigned size)
{
    fwrite(data, size, 1, stderr);

    return 0;
}


static int
copy_output(void *none, const char *data, unsigned size)
{
    fwrite(data, size, 1, stdout);

    return 0;
}


static int
draw_call(void *context, int status, void *output,
	  struct x1f4_dispatch_type **dispatch)
{
    struct express_type *express_data;
    struct fiction_type *fiction_data;

    fiction_data = ((struct context_type *) context)->fiction;

    express_data = fiction_data->express_data;

    status = express_data->call(express_data->back, status, output, dispatch);

    if (*dispatch) {
	struct x1f4_dispatch_type *dispatch_data;

	dispatch_data = *dispatch;

	*dispatch = &express_data->dispatch;

	express_data->back = dispatch_data->back;
	express_data->call = dispatch_data->call;
	express_data->dispatch.transfer = dispatch_data->transfer;
    } else {
	int excess;

	excess = x1f4_fini_shuffle(&express_data->program);
	if (excess) {
	    if (status) {
	    } else {
		status = X1f4_EX_CRITICAL;
	    }
	}

	fiction_data->express_data = express_data->express_data;

	free(express_data);
    }

    return status;
}


static int
ever_list(void *none, void *data, struct x1f4_variable_type *variable_data)
{
    puts(variable_data->name);

    return 0;
}


static int
fini_valist(struct context_type *context_data)
{
#if !__ROLL_VALIST__
# define status                         0
#endif				/* !__ROLL_VALIST__ */

#if __ROLL_VALIST__
    int status = 0;
    void *valist;
#endif				/* __ROLL_VALIST__ */

#if __ROLL_VALIST__
    valist = context_data->valist;
    if (valist) {
	libx1f4i0_fini_valist(context_data->self, valist, 0, 0, &status);
    }
#endif				/* __ROLL_VALIST__ */

    return status;

#if !__ROLL_VALIST__
# undef status
#endif				/* !__ROLL_VALIST__ */
}


static int
fini_xset(struct context_type *context_data)
{
    int excess, status;

    status = x1f4_flat_indexset(context_data->indexset_data);

    excess = fini_valist(context_data);
    if (excess) {
	if (status) {
	} else {
	    status = excess;
	}
    }

    return status;
}


static int
flat_text(void *context)
{
    int status;
    struct context_type *context_data;
    struct x1f4_indexset_type *indexset_data;

    context_data = context;

    indexset_data = context_data->indexset_data;

    status = indexset_data->autodeck_set.deck
	(&indexset_data->autodeck_set.text);

    if (SIGCHLD_received) {
	int excess;

	excess = wait_some(context_data);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
flush_error(struct context_type *context_data, const char *program)
{
    if (context_data->error) {
    } else {
	context_data->error = 1;

	fflush(stdout);
	fprintf(stderr, "%s:", context_data->self);
	if (program) {
	    fprintf(stderr, " %s:", program);
	}
	if (1) {
	    int (*line)(struct context_type *, unsigned *);

	    line = context_data->line;
	    if (line) {
		unsigned copy, miss;

		line(context_data, &copy);
		seek_line(context_data, copy, &miss);
		fprintf(stderr, " %u:", miss + 1);
	    }
	}

	fprintf(stderr, " cannot execute program\n");
    }

    return 17;
}


static int
free_call(void *context, void *data)
{
    void *e2list;

    x1f4_fast_e4ll(e4less(context), &e2list);

    return x1f4_free_e2list(e2list, data);
}


static int
init_valist(struct context_type *context_data)
{
#if __ROLL_VALIST__
    libx1f4i0_init_valist(&context_data->valist, 0, 0);
#endif				/* __ROLL_VALIST__ */

#if __ROLL_VALIST__
    context_data->dlist.ever = 0;
    context_data->dlist.size = 0;
#endif				/* __ROLL_VALIST__ */

#if __ROLL_VALIST__
    context_data->dlist.valist = context_data->valist;
#endif				/* __ROLL_VALIST__ */

    return 0;
}


static int
init_xset(struct context_type *context_data)
{
    int status;
    struct x1f4_textport_type textport;
    unsigned textport_bits = 0;

#if __ROLL_VALIST__
    context_data->valist = NULL;
#endif				/* __ROLL_VALIST__ */

    init_valist(context_data);

    textport_bits = X1f4_TEXTPORT_TEXTFLAT;

    if (context_data->sequenced) {
	textport_bits |= X1f4_TEXTPORT_AUTOLINK;

	textport.autolink_set.data = context_data;
	textport.autolink_set.free = free_call;
	textport.autolink_set.link = link_call;
	textport.autolink_set.mode = mode_call;
	textport.autolink_set.pick = pick_call;
	textport.autolink_set.slip = slip_call;
    }

#if __ROLL_VALIST__
    if (context_data->valist) {
	textport_bits |= X1f4_TEXTPORT_CODELINK;
	libx1f4i0_lead_valist
	    (context_data->valist, 0, 0, &textport.codelink_set,
	     &context_data->dlist);
    }
#endif				/* __ROLL_VALIST__ */

#if __ROLL_VALIST__
    if (context_data->valist) {
	textport_bits |= X1f4_TEXTPORT_RESOURCE;
	libx1f4i0_lead_valist
	    (context_data->valist, 0, 0, &textport.resource_set,
	     &context_data->dlist);
    }
#endif				/* __ROLL_VALIST__ */

    textport.lasttype_set.type = X1f4_E4_LAST;

    textport.tilelong_set.lxtile_data = tiles;
    textport.tilelong_set.miss =
	sizeof(tiles) / sizeof(struct x1f4_lxtile_type);

    lxnear.longpipe_set.datapipe_data = long_pipes;
    lxnear.longpipe_set.miss = 2;

    lxnear.nearpipe_set.textpipe_data = near_pipes;
    lxnear.nearpipe_set.miss = 2;

    textport.textflat_set.data = context_data;
    textport.textflat_set.line = line_flat;
    textport.textflat_set.post = post_flat;
    textport.textflat_set.push = push_flat;

    textport.autodeck_set.class = &context_data->class;

    status = x1f4_fast_indexset
	(context_data->indexset_data, textport_bits, &textport);
    if (status) {
    } else {
	status = rule_text(context_data);
	if (status) {
	    x1f4_flat_indexset(context_data->indexset_data);
	}
    }

    return status;
}


static int
land_fine(void *context, void *output, void **input)
{
    return 0;
}


static int
land_line(void *context, void *output, void **input)
{
    int status;

    do {
	struct context_type *context_data;
	struct express_type *express_data;
	struct x1f4_a1_type a1;
	struct x1f4_a1record_type a1record;
	unsigned flags;
	void *data, *proGram, **reach;

	if (output == x1f4_a1_walk_link) {
	    express_data = malloc(sizeof(struct express_type));
	    if (!express_data) {
		status = -1;
		if (1) {
		    break;
		}
	    }
	} else {
	    express_data = NULL;
	}

	reach = output == x1f4_a1_walk_link ? input[1] : input;

	data = I_TEXT(reach[0]);

	context_data = context;

	posh_line(context_data, &flags, &a1, &a1record);

	status = x1f4_init_shuffle(&proGram, data, flags, &a1);
	if (status) {
	    if (status == X1f4_A1_ALLOC_ERROR) {
		perror(context_data->self);
	    } else {
		line_flat(context_data, -1);
		push_flat(context_data, "cannot parse program: ", 22);
		a1record.pick = 1;
		x1f4_stat_shuffle
		    (stderr, push_flat, data, &a1record,
		     &context_data->indexset_data->eelookup_set.eelookup);
		post_flat(context_data);
	    }

	    status = X1f4_EX_CAN_CONTINUE;
	} else {
	    int excess;

	    if (context_data->sequenced) {
		status = pipe_e4ll(context_data, proGram);
	    } else {
		status = x1f4_pipe_e4ll
		    (context_data->indexset_data->autodeck_set.less, proGram);
	    }
	    if (status) {
		status = X1f4_EX_CRITICAL;
	    } else {
		int args[1];
		struct x1f4_dxcast_type dxcast;
		struct x1f4_function_type function;
		const struct x1f4_linetext_type *linetext_data;
		void **class, *frame[3], *track[3];

		x1f4_hold_shuffle(proGram, &linetext_data);

		x1f4_rail_shuffle(proGram);

		class = reach[-1];

		class[X1f4_E4_RENDITION] = &function;

		track[0] = class;
		track[1] = track + 2;
		track[2] = &dxcast;

		class = track + 1;

		if (output == x1f4_a1_walk_link) {
		    frame[0] = &context_data->record;
		    frame[1] = class;
		    class = frame;
		}

		dxcast.call = 1;
		l_USER((void *) linetext_data, &dxcast.data);

		x1f4_type_e4fine
		    (context_data->indexset_data->transfer_set.fine,
		     &dxcast.type, linetext_data);

		function.args = args;
		function.count = 1;
		function.flags = 0;
		function.type = X1f4_E4_VOID;

		args[0] = X1f4_E4_OBJECT;

		status = context_data->long_call->function.function
		    ((void *) context_data->long_call->context, output, class);
		if (status) {
		} else {
		    if (output == x1f4_a1_walk_link) {
			input[2] = class[2];
		    }
		}
	    }

	    if (status || output != x1f4_a1_walk_link) {
		excess = x1f4_fini_shuffle(&proGram);
		if (excess) {
		    if (status) {
		    } else {
			status = X1f4_EX_CRITICAL;
		    }
		}
	    } else {
		struct x1f4_dispatch_type *dispatch_data;

		express_data->program = proGram;

		dispatch_data = input[2];

		input[2] = &express_data->dispatch;

		express_data->back = dispatch_data->back;
		express_data->call = dispatch_data->call;
		express_data->dispatch.back = context_data;
		express_data->dispatch.call = draw_call;
		express_data->dispatch.transfer = dispatch_data->transfer;
	    }
	}

	if (express_data) {
	    if (status) {
		free(express_data);
	    } else {
		struct fiction_type *fiction_data;

		fiction_data = context_data->fiction;
		express_data->express_data = fiction_data->express_data;
		fiction_data->express_data = express_data;
	    }
	}
    } while (0);

    return status;
}


static int
land_roll(struct context_type *context_data)
{
    int status;
    const void *rall;

    status = x1f4_find_mxpath
	(context_data->indexset_data->function_set.text, "rall", 4, &rall);
    if (status) {
	struct x1f4_function_type *function_data;

	function_data = (void *) rall;

	function_data->function = land_line;

	status = 0;
    } else {
	status = -1;
    }

    return status;
}


static int
lead_less(void **less, struct context_type *context_data)
{
    struct x1f4_e2line_type *e2line_data;
    struct x1f4_e4less_type e4line;

    e2line_data = &context_data->e2line;

    e2line_data->bits = X1f4_E2LIST_FLATLINE;

    e2line_data->e2list.flatline_set.class = &context_data->class;

    e4line.misslink_set.data = e2line_data;

    x1f4_type_e2list(&e4line.misslink_set);

    return x1f4_init_e4ll(less, X1f4_E4LL_MISSLINK, &e4line);
}


static int
line_flat(void *context, int class)
{
    struct context_type *context_data;

    context_data = context;

    fflush(stdout);
    fputs(context_data->self, stderr);
    fputs(":", stderr);
    if (context_data->program) {
	fputs(" ", stderr);
	fputs(context_data->program, stderr);
	fputs(":", stderr);
    }
    if (1) {
	int (*line)(struct context_type *, unsigned *);

	line = context_data->line;
	if (line) {
	    unsigned copy, miss;

	    line(context, &copy);
	    seek_line(context, copy, &miss);
	    fprintf(stderr, " %u:", miss + 1);
	}
    }

    fputs(" ", stderr);

    return 0;
}


static int
line_program(struct context_type *context_data, unsigned *copy)
{
    return x1f4_look_routine(context_data->logic, copy);
}


static int
line_shuffle(struct context_type *context_data, unsigned *copy)
{
    return x1f4_look_shuffle(context_data->logic, copy);
}


static int
link_call(void *context, void **link, unsigned size)
{
    void *e2list;

    x1f4_fast_e4ll(e4less(context), &e2list);

    return x1f4_link_e2list(e2list, link, size);
}


static int
link_e4lf(void *context, void **link, unsigned size)
{
    return x1f4_link_e4lf(e4less(context), link, size);
}


static int
list_text(void *context, const void *text)
{
    x1f4_print_htfunction
	(stdout, text, "", ((struct context_type *) context)->indexset_data,
	 x1f4_look_indexset);

    return 0;
}


static int
lock_e4ll(void *context, const struct x1f4_function_type *function_data)
{
    return x1f4_lock_e4ll(e4less(context), function_data);
}


static int
long_fine(struct context_type *context_data, const char *program)
{
    int status;
    struct fiction_type *fiction_data;

    fiction_data = context_data->fiction_data;

    while (fiction_data) {
	context_data->fiction = fiction_data;
	context_data->logic = fiction_data->part;

	status = x1f4_step_routine(fiction_data->part);
	if (status == X1f4_EX_COMPLETED) {
	    if (1) {
		status = 0;
	    }

	    fiction_data = NULL;
	} else {
	    if (context_data->class) {
		int excess;
		void *e2list;

		context_data->class = 0;

		x1f4_fast_e4ll(e4less(context_data), &e2list);

		excess = x1f4_deck_e2list(&e2list);
		if (excess) {
		    if (status) {
		    } else {
			status = excess;
		    }
		}

		if (SIGCHLD_received) {
		    excess = wait_some(context_data);
		    if (excess) {
			if (status) {
			} else {
			    status = excess;
			}
		    }
		}
	    }

	    if (status) {
		flush_error(context_data, program);
	    }
	}

	if (status) {
	    break;
	}
    }

    return status;
}


static int
long_line(struct context_type *context_data, const char *program, void *degree)
{
    int status;
    struct fiction_type *fiction_data;

    fiction_data = context_data->fiction_data;

    while (fiction_data) {
	context_data->fiction = fiction_data;
	context_data->logic = fiction_data->part;

	status = x1f4_step_shuffle(context_data->logic);
	if (status == X1f4_EX_COMPLETED) {
	    status = 0;

#if 0
	    x1f4_post_shuffle(context_data->logic, degree);
	    if (*(X1f4_E4_C_MODE *) degree) {
		status = *(X1f4_E4_C_MODE *) degree;
	    }
#endif				/* 0 */

	    fiction_data = fiction_data->next_fiction;

	    status = slip_line(context_data, fiction_data->last_fiction);

	    if (!context_data->fiction_data) {
		break;
	    }
	} else {
	    if (context_data->class) {
		int excess;
		void *e2list;

		context_data->class = 0;

		x1f4_fast_e4ll(e4less(context_data), &e2list);

		excess = x1f4_deck_e2list(&e2list);
		if (excess) {
		    if (status) {
		    } else {
			status = excess;
		    }
		}

		if (SIGCHLD_received) {
		    excess = wait_some(context_data);
		    if (excess) {
			if (status) {
			} else {
			    status = excess;
			}
		    }
		}
	    }

	    if (status) {
		flush_error(context_data, program);
	    }

	    fiction_data = fiction_data->next_fiction;
	}

	if (status) {
	    break;
	}
    }

    return status;
}


static int
mode_call(void *context, void **link, unsigned size)
{
    void *e2list;

    x1f4_fast_e4ll(e4less(context), &e2list);

    return x1f4_mode_e2list(e2list, link, size);
}


static int
part_line(struct context_type *context_data, unsigned flags,
	  struct x1f4_a1_type *a1_data)
{
    int status;
    struct x1f4_linetext_type *linetext_data;

    linetext_data = context_data->positive;

    linetext_data->context = context_data;
    linetext_data->function.args = NULL;
    linetext_data->function.count = 0;
    linetext_data->function.flags =
	X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK | X1f4_E4_WALK_LINK;
    linetext_data->function.function = pick_line;
    linetext_data->function.length = 9;
    linetext_data->function.name = "virtual26";
    linetext_data->function.type = X1f4_E4_VOID;

    status = x1f4_post_mxpath
	(context_data->indexset_data->function_set.text, "virtual26", 9,
	 &linetext_data->function);
    if (status) {
    } else {
	status = x1f4_init_shuffle
	    (&context_data->virtual, VIRTUAL, flags, a1_data);
	if (status) {
	} else {
	    status = pipe_e4ll(context_data, context_data->virtual);
	    if (status) {
	    } else {
		x1f4_rail_shuffle(context_data->virtual);
	    }
	}
    }

    return status;
}


static int
past_fork(void *context, void *output, void **input)
{
    int status;
    struct context_type *context_data;
    struct fiction_type *fiction_data, *fiction_lead;

    context_data = context;

    fiction_data = malloc(sizeof(struct fiction_type));
    if (!fiction_data) {
	status = -1;
    } else {
	fiction_lead = context_data->fiction;

	context_data->fiction = fiction_data;

	fiction_data->express_data = NULL;

	status = lead_less(&fiction_data->turn, context_data);
	if (status) {
	} else {
	    context_data->fiction = fiction_data;

	    status = x1f4_near_shuffle
		(&fiction_data->part, context_data->virtual);
	    if (status) {
	    } else {
		int args[2];
		struct x1f4_function_type function, *function_data;
		void **class, **press;

		context_data->fiction = fiction_data;
		context_data->logic = fiction_data->part;

		fiction_data->device = context_data->virtual;

		x1f4_head_shuffle(fiction_data->part);

		class = input;

		press = class[-1];

		function_data = press[X1f4_E4_RENDITION];
		function = *function_data;
		function.args++;
		function.count--;

		press[X1f4_E4_RENDITION] = &function;

		context_data->pick[1] = *class;

		context_data->pick[2] = context_data->pick + 3;

		*class = class[-1];

		status = context_data->long_list->function.function
		    ((void *) context_data->long_list->context,
		     context_data->pick + 3, class + 1);
		if (status) {
		    flush_error(context_data, "dettached");
		} else {
		    *class = context_data->pick[1];

		    function.count = 2;
		    args[0] = function_data->args[0];
		    args[1] = context_data->long_list->function.type;
		    function.args = args;

		    context_data->function_data = &function;

		    context_data->through = 1;

		    do {
			status = x1f4_step_shuffle(context_data->logic);
			if (status) {
			    break;
			} else {
			    if (context_data->class) {
			    }
			}
		    } while (context_data->through);
		}

		press[X1f4_E4_RENDITION] = function_data;

		if (status) {
		    x1f4_side_shuffle
			(&fiction_data->part, context_data->virtual);
		}
	    }

	    context_data->fiction = fiction_lead;

	    if (status) {
		x1f4_fini_e4ll(&fiction_data->turn);
	    }
	}

	if (status) {
	    free(fiction_data);
	} else {
	    struct fiction_type *last_fiction;

	    last_fiction = fiction_lead->last_fiction;

	    last_fiction->next_fiction = fiction_data;

	    fiction_data->last_fiction = last_fiction;
	    fiction_data->next_fiction = fiction_lead;

	    fiction_lead->last_fiction = fiction_data;
	}
    }

    return status;
}


static int
past_line(void *context, void *output, void **input)
{
    return context(context)->long_call->function.function
	((void *) context(context)->long_call->context, output, input);
}


static int
pick_call(void *context, void ***text, int (*miss)(void *), unsigned size,
	  void **link)
{
    void *e2list;

    x1f4_fast_e4ll(e4less(context), &e2list);

    return x1f4_pick_e2list(e2list, text, miss, size, link);
}


static int
pick_e4lf(void *context, void ***text, int (*miss)(void *), unsigned size,
          void **link)
{
    return x1f4_pick_e4lf(e4less(context), text, miss, size, link);
}


static int
pick_line(void *context, void *output, void **class)
{
    int status = 0;
    struct context_type *context_data;

    context_data = context;

    context_data->through = 0;
    class[2] = NULL;
    if (true(output == x1f4_a1_walk_link)) {
	struct fiction_type *fiction_data;
	void **input, **reach;

	fiction_data = context_data->fiction;

	input = fiction_data->input;

	reach = ((void **) class[1])[-1];
	context_data->pick[0] = reach;
	input[0] = &fiction_data->record;
	input[1] = context_data->pick + 1;

	reach[X1f4_E4_RENDITION] = context_data->function_data;

	status = context_data->long_fall->function.function
	    ((void *) context_data->long_fall->context,
	     (void *) x1f4_a1_walk_link, input);
	if (status) {
	} else {
	    class[2] = input[2];
	}
    }

    return status;
}


static int
pick_program(struct context_type *context_data, unsigned *copy)
{
    return x1f4_look_program(context_data->logic, copy);
}


static int
pick_shuffle(struct context_type *context_data, unsigned *copy)
{
    return x1f4_seek_shuffle(context_data->logic, copy);
}


static int
pipe_e4ll(void *context, void *shuffle)
{
    struct x1f4_a1screen_type screen;

    screen.back = lock_e4ll;
    screen.fast = push_e4ll;
    screen.lose = lock_e4ll;

    screen.data = context;

    return x1f4_pipe_shuffle(shuffle, &screen);
}


static int
posh_line(struct context_type *context_data, unsigned *flags,
	  struct x1f4_a1_type *a1_data,
	  struct x1f4_a1record_type *a1record_data)
{
    x1f4_xset_shuffle(a1_data, flags, context_data->indexset_data);

    *flags |= X1f4_A1_BCOLLECT | X1f4_A1_COMPOSER | X1f4_A1_SIDELIST;

    if (context_data->sequenced) {
	a1_data->autolead_set.context = context_data;
	a1_data->autolead_set.link = link_e4lf;
	a1_data->autolead_set.pick = pick_e4lf;
    }

    a1_data->bcollect_set.a1record_data = a1record_data;
    set_composer(context_data, &a1_data->composer_set.context,
		 &a1_data->composer_set.get);
    x1f4_pset_lxcast
	((void *) a1_data->composer_set.context,
	 (void *) &a1_data->sidetype_set.miss);

#if __ROLL_VALIST__
    if (context_data->valist) {
	*flags |= X1f4_A1_RESOURCE;
	libx1f4i0_lead_valist
	    (context_data->valist, 0, 0, &a1_data->resource_set,
	     &context_data->dlist);
    }
#endif				/* __ROLL_VALIST__ */

    return 0;
}


static int
post_flat(void *context)
{
    ((struct context_type *) context)->error = 1;

    fputs("\n", stderr);

    return 0;
}


static int
push_case(struct context_type *context_data, const char *program,
	  const void *data)
{
    int status;

    if (x1f4_peer_shuffle(data, X1f4_A1_SCOMMENT)) {
	status = push_line(context_data, program, data);
    } else {
	status = push_fine(context_data, program, data);
    }

    return status;
}


static int
push_e4ll(void *context, const struct x1f4_function_type *function_data)
{
    return x1f4_push_e4ll(e4less(context), function_data);
}


static int
push_fine(struct context_type *context_data, const char *program,
	  const void *data)
{
    int status;
    struct x1f4_c1_type c1;
    struct x1f4_c1record_type c1record;
    unsigned flags;
    void *proGram;

    x1f4_xset_program(&c1, &flags, context_data->indexset_data);

    flags |= X1f4_C1_BCOLLECT | X1f4_C1_COMPOSER | X1f4_C1_SIDELIST;

    c1.bcollect_set.c1record_data = &c1record;
    set_composer(context_data, &c1.composer_set.context, &c1.composer_set.get);
    x1f4_pset_lxcast
	((void *) c1.composer_set.context, (void *) &c1.sidetype_set.miss);

#if __ROLL_VALIST__
    if (context_data->valist) {
	flags |= X1f4_C1_RESOURCE;
	libx1f4i0_lead_valist
	    (context_data->valist, 0, 0, &c1.resource_set,
	     &context_data->dlist);
    }
#endif				/* __ROLL_VALIST__ */

    status = x1f4_init_program(&proGram, data, flags, &c1);
    if (status) {
	if (status == X1f4_C1_ALLOC_ERROR) {
	    perror(context_data->self);
	} else {
	    fprintf(stderr, "%s: cannot parse ", context_data->self);
	    if (program) {
		fprintf(stderr, "`%s'", program);
	    } else {
		fprintf(stderr, "program");
	    }
	    fprintf(stderr, "\n");
	    fprintf(stderr, "%s: ", context_data->self);
	    c1record.pick = 1;
	    x1f4_stat_program
		(stderr, test_cast, data, &c1record,
		 &context_data->indexset_data->eelookup_set.eelookup);
	    fprintf(stderr, "\n");
	}
    } else {
	back_roll(context_data);

	if (0) {
	} else {
	    context_data->class = 0;
	    context_data->error = 0;

	    context_data->fork = 0;

	    context_data->device = proGram;
	    context_data->logic = proGram;

	    static_class = &context_data->class;

	    if (context_data->sequenced) {
		status = roll_fine(context_data, program, proGram);
	    } else {
		context_data->line = pick_program;

		status = x1f4_note_program
		    (proGram, &context_data->class, context_data, flat_text);
		if (status) {
		    flush_error(context_data, program);
		}
	    }
	}

	x1f4_fini_program(&proGram);
    }

    return status;
}


static int
push_flat(void *context, const char *data, unsigned size)
{
    fwrite(data, size, 1, stderr);

    return 0;
}


static int
push_line(struct context_type *context_data, const char *program,
	  const void *data)
{
    int status;
    struct x1f4_a1_type a1;
    struct x1f4_a1record_type a1record;
    unsigned flags;
    void *proGram;

    if (context_data->sequenced) {
	roll_fork(context_data);
    }
    {
	land_roll(context_data);
    }

    posh_line(context_data, &flags, &a1, &a1record);

    status = x1f4_init_shuffle(&proGram, data, flags, &a1);
    if (status) {
	if (status == X1f4_A1_ALLOC_ERROR) {
	    perror(context_data->self);
	} else {
	    fprintf(stderr, "%s: cannot parse ", context_data->self);
	    if (program) {
		fprintf(stderr, "`%s'", program);
	    } else {
		fprintf(stderr, "program");
	    }
	    fprintf(stderr, "\n");
	    fprintf(stderr, "%s: ", context_data->self);
	    a1record.pick = 1;
	    x1f4_stat_shuffle
		(stderr, test_cast, data, &a1record,
		 &context_data->indexset_data->eelookup_set.eelookup);
	    fprintf(stderr, "\n");
	}
    } else {
	int excess;

	back_roll(context_data);

	context_data->virtual = NULL;

	if (context_data->sequenced) {
	    status = part_line(context_data, flags, &a1);
	    if (status) {
	    } else {
		status = pipe_e4ll(context_data, proGram);
	    }
	} else {
	    status = x1f4_pipe_e4ll
		(context_data->indexset_data->autodeck_set.less, proGram);
	}
	if (status) {
	    perror(context_data->self);
	} else {
	    X1f4_E4_C_MODE degree;

	    context_data->class = 0;
	    context_data->error = 0;

	    context_data->fork = 1;

	    context_data->device = proGram;
	    context_data->logic = proGram;

	    static_class = &context_data->class;

	    if (context_data->sequenced) {
		status = roll_line(context_data, program, proGram, &degree);
	    } else {
		context_data->line = pick_shuffle;

		status = x1f4_long_shuffle
		    (proGram, &context_data->class, context_data,
		     flat_text, &degree);
		if (status) {
		    flush_error(context_data, program);
		} else {
		    status = degree;
		}
	    }
	}

	if (context_data->virtual) {
	    excess = x1f4_fini_shuffle(&context_data->virtual);
	    if (excess) {
		if (status) {
		} else {
		    status = excess;
		}
	    }
	}

	excess = x1f4_fini_shuffle(&proGram);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
push_port(struct context_type *context_data, int argc, char **argv)
{
    int status;
    struct x1f4_indexset_type indexset;

    context_data->indexset_data = &indexset;

    status = init_xset(context_data);
    if (status) {
    } else {
	const char *program;

	program = argv[optind];
	optind++;

	SIGCHLD_received = 0;

	signal(SIGCHLD, SIGCHLD_handler);

	if (context_data->immediate) {
	    if (0) {
	    } else {
		context_data->data = program;

		context_data->program = NULL;

		status = push_case(context_data, NULL, program);
	    }
	} else {
	    unsigned size;
	    void *data;

	    status = libx1f4i0_read_file(&data, &size, program, 1);
	    if (status) {
		libx1f4i0_stat_failure(argv[0], status, program);
	    } else {
		if (1) {
		    char *mine;

		    mine = data;
		    mine[size] = 0;
		}

		context_data->data = data;

		context_data->program = program;

		status = push_case(context_data, program, data);

		free(data);
	    }
	}

	signal(SIGCHLD, SIG_DFL);

	if (1) {
	    int excess;

	    excess = fini_xset(context_data);
	    if (excess) {
		if (status) {
		} else {
		    status = excess;
		}
	    }
	}
    }

    return status;
}


static int
roll_fine(struct context_type *context_data, const char *program,
	  void *proGram)
{
    int status;
    struct fiction_type *fiction_data;

    x1f4_rail_program(proGram);

    context_data->line = line_program;

    fiction_data = malloc(sizeof(struct fiction_type));
    if (!fiction_data) {
	status = -1;
    } else {
	int excess;

	context_data->fiction_data = fiction_data;

	fiction_data->last_fiction = fiction_data;
	fiction_data->next_fiction = fiction_data;

	fiction_data->express_data = NULL;

	fiction_data->part = NULL;
	fiction_data->turn = NULL;

	status = lead_less(&fiction_data->turn, context_data);
	if (status) {
	} else {
	    context_data->fiction = fiction_data;

	    status = x1f4_near_routine(&fiction_data->part, proGram);
	    if (status) {
	    } else {
		x1f4_head_routine(fiction_data->part);
	    }

	    if (status) {
	    } else {
		status = long_fine(context_data, program);
	    }
	}

	/* one `fiction' program in the list */

	fiction_data = context_data->fiction_data;
	if (fiction_data) {
	    struct fiction_type *fiction_lead;

	    fiction_lead = fiction_data;

	    do {
		context_data->fiction = fiction_data;

		if (fiction_data->part) {
		    excess = x1f4_side_routine(&fiction_data->part, proGram);
		    if (excess) {
			if (status) {
			} else {
			    status = excess;
			}
		    }
		}

		if (fiction_data->turn) {
		    excess = x1f4_fini_e4ll(&fiction_data->turn);
		    if (excess) {
			if (status) {
			} else {
			    status = excess;
			}
		    }
		}

		fiction_data = fiction_data->next_fiction;

		free(fiction_data->last_fiction);
	    } while (fiction_data != fiction_lead);
	}
    }

    return status;
}


static int
roll_fork(struct context_type *context_data)
{
    int status;
    const void *pall;

    status = x1f4_find_mxpath
	(context_data->indexset_data->function_set.text, "pall", 4, &pall);
    if (status) {
	struct x1f4_function_type *function_data;

	function_data = (void *) pall;

	function_data->function = past_fork;

	status = 0;
    } else {
	status = -1;
    }

    return status;
}


static int
roll_line(struct context_type *context_data, const char *program,
	  void *proGram, void *degree)
{
    int status;
    struct fiction_type *fiction_data;

    x1f4_rail_shuffle(proGram);

    context_data->line = line_shuffle;

    fiction_data = malloc(sizeof(struct fiction_type));
    if (!fiction_data) {
	status = -1;
    } else {
	int excess;

	context_data->fiction_data = fiction_data;

	fiction_data->last_fiction = fiction_data;
	fiction_data->next_fiction = fiction_data;

	fiction_data->express_data = NULL;

	fiction_data->part = NULL;
	fiction_data->turn = NULL;

	status = lead_less(&fiction_data->turn, context_data);
	if (status) {
	} else {
	    context_data->fiction = fiction_data;

	    status = x1f4_near_shuffle(&fiction_data->part, proGram);
	    if (status) {
	    } else {
		fiction_data->device = proGram;

		x1f4_head_shuffle(fiction_data->part);
	    }

	    if (status) {
	    } else {
		status = long_line(context_data, program, degree);
	    }
	}

	fiction_data = context_data->fiction_data;
	if (fiction_data) {
	    struct fiction_type *fiction_lead;

	    fiction_lead = fiction_data;

	    do {
		struct fiction_type *fiction_miss;

		context_data->fiction = fiction_data;

		if (fiction_data->part) {
		    excess = x1f4_side_shuffle
			(&fiction_data->part, fiction_data->device);
		    if (excess) {
			if (status) {
			} else {
			    status = excess;
			}
		    }
		}

		if (fiction_data->turn) {
		    excess = x1f4_fini_e4ll(&fiction_data->turn);
		    if (excess) {
			if (status) {
			} else {
			    status = excess;
			}
		    }
		}

		fiction_miss = fiction_data;

		fiction_data = fiction_data->next_fiction;

		free(fiction_miss);
	    } while (fiction_data != fiction_lead);
	}
    }

    return status;
}


static int
rule_fset(struct context_type *context_data,
	  const struct x1f4_linetext_type *linetext_data)
{
    const char *name;
    int status;
    void *fsdeck;

    fsdeck = context_data->indexset_data->function_set.text;

    name = linetext_data->function.name;
    if (name) {
	do {
	    status = x1f4_post_mxpath
		(fsdeck, name, linetext_data->function.length,
		 &linetext_data->function);
	    if (status) {
		break;
	    } else {
		linetext_data++;
	    }

	    name = linetext_data->function.name;
	} while (name);
    } else {
	status = 0;
    }

    if (status) {
    } else {
	const void *long_call, *long_fall, *long_list;

	status = x1f4_find_mxpath(fsdeck, "l_assemble", 10, &long_list);
	if (status) {
	    status = x1f4_find_mxpath(fsdeck, "l_xcall", 7, &long_fall);
	    if (status) {
		status = x1f4_find_mxpath(fsdeck, "call", 4, &long_call);
		if (status) {
		    status = 0;

		    context_data->long_call =
			(const void *)
			((const char *) long_call
			 - offsetof(struct x1f4_linetext_type, function));
		    context_data->long_fall =
			(const void *)
			((const char *) long_fall
			 - offsetof(struct x1f4_linetext_type, function));
		    context_data->long_list =
			(const void *)
			((const char *) long_list
			 - offsetof(struct x1f4_linetext_type, function));
		} else {
		    status = -1;
		}
	    } else {
		status = -1;
	    }
	} else {
	    status = -1;
	}
    }

    return status;
}


static int
rule_text(struct context_type *context_data)
{
    struct x1f4_linetext_type *linetext_data;

    linetext_data = context_data->linetext;

    linetext_data->context = context_data;
    linetext_data->function.args = &used_object;
    linetext_data->function.count = 1;
    linetext_data->function.flags =
	X1f4_E4_KEEP_CALL | X1f4_E4_SIDE_LIST | X1f4_E4_TEXT_LINK;
    linetext_data->function.function = past_line;
    linetext_data->function.length = 4;
    linetext_data->function.name = "pall";
    linetext_data->function.type = X1f4_E4_VOID;

    linetext_data++;

    linetext_data->context = context_data;
    linetext_data->function.args = &used_text;
    linetext_data->function.count = 1;
    linetext_data->function.flags =
	X1f4_E4_KEEP_CALL | X1f4_E4_TEXT_LINK | X1f4_E4_WALK_LINK;
    linetext_data->function.function = land_fine;
    linetext_data->function.length = 4;
    linetext_data->function.name = "rall";
    linetext_data->function.type = X1f4_E4_VOID;

    linetext_data++;

    linetext_data->function.name = NULL;

    return rule_fset(context_data, linetext_data - 2);
}


static int
seek_line(struct context_type *context_data, unsigned copy, unsigned *miss)
{
    const char *data;
    unsigned line = 0;

    data = context_data->data;
    while (copy) {
	if (*data == 10) {
	    line++;
	}

	copy--;
	data++;
    }

    *miss = line;

    return 0;
}


static int
set_composer(struct context_type *context_data, void **text,
	     int (**lock)(void *, int, int,
			  const struct x1f4_linetext_type **))
{
    struct x1f4_lxslip_type *lxslip_data;

    lxslip_data = context_data->indexset_data->sliplong_set.lxslip_data;
    while (lxslip_data->note != x1f4_note_lxcast) {
	lxslip_data++;
    }

    *lock = x1f4_lock_lxcast;

    *text = lxslip_data->slip;

    return 0;
}


static int
slip_call(void *context, void *data, void *text, int (*miss)(void *))
{
    void *e2list;

    x1f4_fast_e4ll(e4less(context), &e2list);

    return x1f4_slip_e2list(e2list, data, text, miss);
}


static int
slip_line(struct context_type *context_data, struct fiction_type *fiction_data)
{
    int excess, status;
    struct fiction_type *last_fiction, *next_fiction;

    status = x1f4_side_shuffle(&fiction_data->part, fiction_data->device);

    excess = x1f4_fini_e4ll(&fiction_data->turn);
    if (excess) {
	if (status) {
	} else {
	    status = excess;
	}
    }

    last_fiction = fiction_data->last_fiction;
    next_fiction = fiction_data->next_fiction;

    if (next_fiction != fiction_data) {
	last_fiction->next_fiction = next_fiction;
	next_fiction->last_fiction = last_fiction;

	context_data->fiction_data = next_fiction;
    } else {
	context_data->fiction_data = NULL;
    }

    free(fiction_data);

    return status;
}


static int
test_cast(void *cast, const char *name, unsigned size)
{
    return size ^ fwrite(name, 1, size, cast);
}


static int
wait_some(struct context_type *context_data)
{
    int status;
    struct x1f4_lxslip_type *lxslip_data;
    void *lxport;

    SIGCHLD_received = 0;

    lxslip_data = context_data->indexset_data->sliplong_set.lxslip_data;
    while (lxslip_data->note != x1f4_note_lxport) {
	lxslip_data++;
    }

    lxport = lxslip_data->slip;

    while (1) {
	pid_t wait;

	wait = waitpid(-1, NULL, WNOHANG);
	if (wait == -1) {
	    status = 0;
	    if (1) {
		break;
	    }
	} else {
	    if (wait) {
		x1f4_wait_lxport(lxport, wait);
	    } else {
		status = 0;
		if (1) {
		    break;
		}
	    }
	}
    }

    return status;
}


static void
SIGCHLD_handler(int signal)
{
    SIGCHLD_received = 1;
    *static_class = 1;
}


static void
list_0010(struct context_type *context_data)
{
    struct x1f4_indexset_type indexset;

    context_data->indexset_data = &indexset;

    if (init_xset(context_data)) {
    } else {
	x1f4_lime_mxdeck(indexset.function_set.text, context_data, list_text);

	fini_xset(context_data);
    }
}


static void
list_0220(struct context_type *context_data)
{
    struct x1f4_indexset_type indexset;

    context_data->indexset_data = &indexset;

    if (init_xset(context_data)) {
    } else {
	x1f4_list_state(indexset.variable_set.text, NULL, ever_list);

	fini_xset(context_data);
    }
}


static void
usage(void)
{
    puts("Usage: tf26 [OPTIONS] PROGRAM\n\
Execute PROGRAM.\n\
\n\
Options:\n\
  -c				execute the PROGRAM program instead program\n\
				read from the PROGRAM file\n\
  -f, --framed			run function calls in a hierarchical fashion\n\
  -s, --sequenced		flatten call hierarchy\n\
      --data			list defined constants and exit\n\
      --list			list available functions and exit\n\
      --help			display this help and exit\n\
      --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int list_functions = 0, status = 0;
    struct context_type context;

    context.immediate = 0;

    context.sequenced = 0;

    unsetenv("POSIXLY_CORRECT");

    {
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "data",	    0x00,   NULL,   0x02    },
		{   "framed",	    0x00,   NULL,   0x66    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   "list",	    0x00,   NULL,   0x01    },
		{   "print",	    0x00,   NULL,   0x50    },
		{   "sequenced",    0x00,   NULL,   0x73    },
		{   "version",	    0x00,   NULL,   0x76    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "fcs", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 001:
		list_functions = 1;
		break;
	    case 002:
		list_functions = 2;
		break;
	    case 'c':
		context.immediate = 1;
		break;
	    case 'f':
		context.sequenced = 0;
		break;
	    case 'h':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case 's':
		context.sequenced = 1;
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}
    }

    if (list_functions) {
	if (list_functions == 1) {
	    list_0010(&context);
	} else {
	    if (list_functions == 2) {
		list_0220(&context);
	    }
	}
    } else {
	if (argc - optind != 1) {
	    fprintf(stderr, "%s: wrong number of arguments\nTry `%s --help' "
		    "for more information.\n", argv[0], argv[0]);

	    return 1;
	} else {
	    context.self = argv[0];

	    status = push_port(&context, argc, argv);
	}
    }

    return status;
}
